from __future__ import annotations

import logging
import threading
import urllib.parse
from contextlib import contextmanager
from typing import ClassVar, Optional, Dict, Iterator, List, Tuple, Any, Set, Sequence

import pandas as pd
import sqlalchemy
from sqlalchemy.exc import ProgrammingError

from metricflow.protocols.sql_client import SqlEngine, SqlEngineAttributes
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.common_client import SqlDialect, not_empty
from metricflow.sql_clients.sqlalchemy_dialect import SqlAlchemySqlClient


class SnowflakeEngineAttributes(SqlEngineAttributes):
    """Engine-specific attributes for the Snowflake query engine

    This is an implementation of the SqlEngineAttributes protocol for Snowflake
    """

    sql_engine_type: ClassVar[SqlEngine] = SqlEngine.SNOWFLAKE

    # SQL Engine capabilities
    date_trunc_supported: ClassVar[bool] = True
    full_outer_joins_supported: ClassVar[bool] = True
    indexes_supported: ClassVar[bool] = False
    multi_threading_supported: ClassVar[bool] = True
    timestamp_type_supported: ClassVar[bool] = True
    timestamp_to_string_comparison_supported: ClassVar[bool] = True
    cancel_submitted_queries_supported: ClassVar[bool] = True

    # SQL Dialect replacement strings
    double_data_type_name: ClassVar[str] = "DOUBLE"
    timestamp_type_name: ClassVar[Optional[str]] = "TIMESTAMP"

    # MetricFlow attributes
    sql_query_plan_renderer: ClassVar[SqlQueryPlanRenderer] = DefaultSqlQueryPlanRenderer()


logger = logging.getLogger(__name__)


class SnowflakeSqlClient(SqlAlchemySqlClient):
    """Client for Snowflake.

    Note: By default, Snowflake uses uppercase for schema, table, and column
    names. To create or access them as lowercase, you must use double quotes.

    It's also tricky trying to get tests / queries on Snowflake working with
    https://docs.snowflake.com/en/sql-reference/parameters.html#quoted-identifiers-ignore-case enabled.
    For example, when listing table names, all tables would be upper case with that setting (causing an issue where
    data sources would constantly be primed because the table names didn't match).
    """

    DEFAULT_LOGIN_TIMEOUT = 60
    DEFAULT_CLIENT_SESSION_KEEP_ALIVE = True

    @staticmethod
    def _parse_url_query_params(url: str) -> Dict[str, str]:
        """Gets the warehouse from the query parameters in the URL, throwing an exception if not set properly."""
        url_query_params: Dict[str, str] = {}

        parsed_url = urllib.parse.urlparse(url)
        query_dict = urllib.parse.parse_qs(parsed_url.query)

        if "warehouse" not in query_dict:
            raise ValueError(f"Missing warehouse in URL query: {url}")

        if len(query_dict["warehouse"]) > 1:
            raise ValueError(f"Multiple warehouses in URL query: {url}")

        url_query_params["warehouse"] = query_dict["warehouse"][0]

        # optionally, role
        if "role" not in query_dict:
            return url_query_params

        if len(query_dict["role"]) > 1:
            raise ValueError(f"Multiple roles in URL query: {url}")

        url_query_params["role"] = query_dict["role"][0]
        return url_query_params

    @staticmethod
    def from_connection_details(url: str, password: Optional[str]) -> SnowflakeSqlClient:  # noqa: D
        parsed_url = sqlalchemy.engine.make_url(url)
        if parsed_url.drivername != SqlDialect.SNOWFLAKE.value:
            raise ValueError(f"Invalid dialect in URL for Snowflake: {url}")

        if parsed_url.port:
            raise ValueError(f"Snowflake URL should not have a port set: {url}")

        if not password:
            raise ValueError(f"Password not supplied for {url}")

        SqlAlchemySqlClient.validate_query_params(
            url=parsed_url, required_parameters={"warehouse"}, optional_parameters={"role"}
        )

        return SnowflakeSqlClient(
            host=not_empty(parsed_url.host, "host", url),
            username=not_empty(parsed_url.username, "username", url),
            password=password,
            database=not_empty(parsed_url.database, "database", url),
            url_query_params=SnowflakeSqlClient._parse_url_query_params(url),
        )

    def __init__(  # noqa: D
        self,
        database: str,
        username: str,
        password: str,
        host: str,
        url_query_params: Dict[str, str],
        login_timeout: int = DEFAULT_LOGIN_TIMEOUT,
        client_session_keep_alive: bool = DEFAULT_CLIENT_SESSION_KEEP_ALIVE,
    ) -> None:
        self._connection_url = SqlAlchemySqlClient.build_engine_url(
            dialect=SqlDialect.SNOWFLAKE.value,
            username=username,
            password=password,
            host=host,
            database=database,
            query=url_query_params,
        )
        self._engine_lock = threading.Lock()
        self._known_sessions_ids_lock = threading.Lock()
        self._known_session_ids: Set[int] = set()
        super().__init__(
            engine=self._create_engine(login_timeout=login_timeout, client_session_keep_alive=client_session_keep_alive)
        )

    def _create_engine(
        self,
        login_timeout: int = DEFAULT_LOGIN_TIMEOUT,
        client_session_keep_alive: bool = DEFAULT_CLIENT_SESSION_KEEP_ALIVE,
    ) -> sqlalchemy.engine.Engine:  # noqa: D
        return sqlalchemy.create_engine(
            self._connection_url,
            pool_size=10,
            max_overflow=10,
            pool_pre_ping=False,
            connect_args={"client_session_keep_alive": client_session_keep_alive, "login_timeout": login_timeout},
        )

    @property
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Collection of attributes and features specific to the Snowflake SQL engine"""
        return SnowflakeEngineAttributes()

    @contextmanager
    def engine_connection(self, engine: sqlalchemy.engine.Engine) -> Iterator[sqlalchemy.engine.Connection]:
        """Context Manager for providing a configured connection.

        Snowflake allows setting a WEEK_START parameter on each session. This forces the value to be
        1, which means Monday. Future updates could parameterize this to read from some kind of
        options construct, which the DBClient could read in at initialization and use here (for example).
        At this time we hard-code the ISO standard.
        """

        with super().engine_connection(self._engine) as conn:
            # WEEK_START 1 means Monday.
            conn.execute("ALTER SESSION SET WEEK_START = 1;")
            results = conn.execute("SELECT CURRENT_SESSION()")
            sessions = []
            for row in results:
                sessions.append(row[0])
            assert len(sessions) == 1
            session = sessions[0]
            with self._known_sessions_ids_lock:
                self._known_session_ids.add(session)
            yield conn
            with self._known_sessions_ids_lock:
                self._known_session_ids.remove(session)

    def _query(  # noqa: D
        self,
        stmt: str,
        bind_params: SqlBindParameters = SqlBindParameters(),
        allow_re_auth: bool = True,
    ) -> pd.DataFrame:
        with self.engine_connection(engine=self._engine) as conn:
            try:
                return pd.read_sql_query(sqlalchemy.text(stmt), conn, params=bind_params.param_dict)
            except ProgrammingError as e:
                if "Authentication token has expired" in str(e) and allow_re_auth:
                    logger.warning(
                        "Snowflake authentication token expired. Attempting to re-auth, then we'll re-run the query"
                    )
                    with self._engine_lock:
                        self._engine.dispose()
                        self._engine = self._create_engine()
                    # this was our one chance to re-auth
                    return self._query(stmt, allow_re_auth=False, bind_params=bind_params)
                raise e

    def _engine_specific_query_implementation(  # noqa: D
        self, stmt: str, bind_params: SqlBindParameters
    ) -> pd.DataFrame:
        return self._query(stmt, bind_params)

    def list_tables(self, schema_name: str) -> Sequence[str]:  # noqa: D
        df = self.query(
            f"SHOW TABLES IN {schema_name}",
        )
        if df.empty:
            return []

        # Lower casing table names to be similar to other SQL clients. TBD on the implications of this.
        return [t.lower() for t in df["name"]]

    def generate_health_check_tests(self, schema_name: str) -> List[Tuple[str, Any]]:  # type: ignore # noqa: D
        additional_tests = [
            (
                "Connection State",
                lambda: str(
                    self.query(
                        "SELECT CURRENT_USER(), CURRENT_ROLE(), CURRENT_DATABASE(), CURRENT_WAREHOUSE(), CURRENT_SCHEMA();"
                    )
                ),
            )
        ]
        return super().generate_health_check_tests(schema_name=schema_name) + additional_tests

    def close(self) -> None:
        """Snowflake will hang pytest if this is not done."""
        with self._engine_lock:
            self._engine.dispose()

    def cancel_submitted_queries(self) -> None:  # noqa: D
        with super().engine_connection(self._engine) as conn:
            with self._known_sessions_ids_lock:
                for session_id in self._known_session_ids:
                    logger.info(f"Cancelling queries associated with session id: {session_id}")
                    conn.execute(f"SELECT SYSTEM$cancel_all_queries({session_id})")



import logging
from typing import ClassVar, Optional, Mapping, Union, Sequence

import sqlalchemy

from metricflow.protocols.sql_client import SqlEngine, SqlEngineAttributes
from metricflow.sql.render.redshift import RedshiftSqlQueryPlanRenderer
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql_clients.common_client import SqlDialect, not_empty
from metricflow.sql_clients.sqlalchemy_dialect import SqlAlchemySqlClient

logger = logging.getLogger(__name__)


class RedshiftEngineAttributes(SqlEngineAttributes):
    """Engine-specific attributes for the Redshift query engine

    This is an implementation of the SqlEngineAttributes protocol for Redshift
    """

    sql_engine_type: ClassVar[SqlEngine] = SqlEngine.REDSHIFT

    # SQL Engine capabilities
    date_trunc_supported: ClassVar[bool] = True
    full_outer_joins_supported: ClassVar[bool] = True
    indexes_supported: ClassVar[bool] = False
    multi_threading_supported: ClassVar[bool] = True
    timestamp_type_supported: ClassVar[bool] = True
    timestamp_to_string_comparison_supported: ClassVar[bool] = True
    # Cancelling should be possible, but not yet implemented.
    cancel_submitted_queries_supported: ClassVar[bool] = False

    # SQL Dialect replacement strings
    double_data_type_name: ClassVar[str] = "DOUBLE PRECISION"
    timestamp_type_name: ClassVar[Optional[str]] = "TIMESTAMP"

    # MetricFlow attributes
    sql_query_plan_renderer: ClassVar[SqlQueryPlanRenderer] = RedshiftSqlQueryPlanRenderer()


class RedshiftSqlClient(SqlAlchemySqlClient):
    """Implements Redshift."""

    @staticmethod
    def from_connection_details(url: str, password: Optional[str]) -> SqlAlchemySqlClient:  # noqa: D
        parsed_url = sqlalchemy.engine.url.make_url(url)
        dialect = SqlDialect.REDSHIFT.value
        if parsed_url.drivername != dialect:
            raise ValueError(f"Expected dialect '{dialect}' in {url}")

        if password is None:
            raise ValueError(f"Password not supplied for {url}")

        return RedshiftSqlClient(
            host=not_empty(parsed_url.host, "host", url),
            port=not_empty(parsed_url.port, "port", url),
            username=not_empty(parsed_url.username, "username", url),
            password=password,
            database=not_empty(parsed_url.database, "database", url),
            query=parsed_url.query,
        )

    def __init__(  # noqa: D
        self,
        port: int,
        database: str,
        username: str,
        password: str,
        host: str,
        query: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
    ) -> None:
        super().__init__(
            engine=self.create_engine(
                dialect=SqlDialect.REDSHIFT.value,
                driver="psycopg2",
                port=port,
                database=database,
                username=username,
                password=password,
                host=host,
                query=query,
            )
        )

    @property
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Collection of attributes and features specific to the Snowflake SQL engine"""
        return RedshiftEngineAttributes()

    def cancel_submitted_queries(self) -> None:  # noqa: D
        raise NotImplementedError



from metricflow.dataflow.sql_column import SqlColumn
from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.context.data_warehouse import (
    ColumnProperties,
    InferenceColumnType,
    DataWarehouseInferenceContext,
    TableProperties,
)


def test_column_properties_is_empty() -> None:
    """Just some easy assertions to test is_empty works as intended."""
    props = ColumnProperties(
        column=SqlColumn.from_string("db.schema.table.column"),
        type=InferenceColumnType.INTEGER,
        row_count=10000,
        distinct_row_count=1000,
        is_nullable=True,
        null_count=1,
        min_value=0,
        max_value=9999,
    )
    assert not props.is_empty

    empty_props = ColumnProperties(
        column=SqlColumn.from_string("db.schema.table.column"),
        type=InferenceColumnType.UNKNOWN,
        row_count=0,
        distinct_row_count=0,
        is_nullable=False,
        null_count=0,
        min_value=None,
        max_value=None,
    )
    assert empty_props.is_empty


def test_table_properties() -> None:
    """Test `TableProperties` initialization.

    This test case asserts that the conversion from the `column_props` argument (which is a list) to
    `self.columns` (which is a dict) implemented by `TableProperties.__post_init__` works as intended.
    """
    table = SqlTable.from_string("db.schema.table")
    col_props = [
        ColumnProperties(
            column=SqlColumn(table=table, column_name="column1"),
            type=InferenceColumnType.INTEGER,
            row_count=1000,
            distinct_row_count=1000,
            is_nullable=False,
            null_count=0,
            min_value=0,
            max_value=999,
        ),
        ColumnProperties(
            column=SqlColumn(table=table, column_name="column2"),
            type=InferenceColumnType.FLOAT,
            row_count=2000,
            distinct_row_count=1000,
            is_nullable=True,
            null_count=10,
            min_value=0,
            max_value=1000,
        ),
    ]

    table_props = TableProperties(table=table, column_props=col_props)

    assert table_props.columns == {
        col_props[0].column: col_props[0],
        col_props[1].column: col_props[1],
    }


def test_data_warehouse_inference_context() -> None:
    """Test `DataWarehouseInferenceContext` initialization.

    This test case asserts that the conversion from the `table_props` argument
    (which is a list) to `self.tables` and `self.columns` (which are dicts)
    implemented by `DataWarehouseInferenceContext.__post_init__` works as intended.
    """
    t1 = SqlTable.from_string("db.schema1.table1")
    t2 = SqlTable.from_string("db.schema2.table1")

    t1_cols = [
        ColumnProperties(
            column=SqlColumn(table=t1, column_name="column1"),
            type=InferenceColumnType.INTEGER,
            row_count=1000,
            distinct_row_count=1000,
            is_nullable=False,
            null_count=0,
            min_value=0,
            max_value=999,
        ),
        ColumnProperties(
            column=SqlColumn(table=t1, column_name="column2"),
            type=InferenceColumnType.FLOAT,
            row_count=2000,
            distinct_row_count=1000,
            is_nullable=True,
            null_count=10,
            min_value=0,
            max_value=1000,
        ),
    ]
    t1_props = TableProperties(table=t1, column_props=t1_cols)

    t2_cols = [
        ColumnProperties(
            column=SqlColumn(table=t2, column_name="column_a"),
            type=InferenceColumnType.FLOAT,
            row_count=1000,
            distinct_row_count=1000,
            is_nullable=False,
            null_count=0,
            min_value=0,
            max_value=999,
        ),
    ]
    t2_props = TableProperties(table=t2, column_props=t2_cols)

    ctx = DataWarehouseInferenceContext(table_props=[t1_props, t2_props])

    assert ctx.tables == {t1: t1_props, t2: t2_props}

    assert ctx.columns == {
        t1_cols[0].column: t1_cols[0],
        t1_cols[1].column: t1_cols[1],
        t2_cols[0].column: t2_cols[0],
    }



from metricflow.dataflow.sql_column import SqlColumn
from metricflow.inference.models import InferenceSignal, InferenceSignalConfidence, InferenceSignalType
from metricflow.inference.solver.weighted_tree import WeightedTypeTreeInferenceSolver

column = SqlColumn.from_string("db.schema.table.col")
solver = WeightedTypeTreeInferenceSolver()


def test_empty_signals_return_unknown() -> None:  # noqa: D
    result = solver.solve_column(column, [])

    assert result.type_node == InferenceSignalType.UNKNOWN
    assert len(result.reasons) == 0
    assert len(result.problems) == 2


def test_follow_signal_path() -> None:
    """Test that the solver will return the deepest (most specific) node if it finds a path with multiple signals."""
    signals = [
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.UNIQUE,
            reason="UNIQUE",
            confidence=InferenceSignalConfidence.HIGH,
            only_applies_to_parent=False,
        ),
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.PRIMARY,
            reason="PRIMARY",
            confidence=InferenceSignalConfidence.VERY_HIGH,
            only_applies_to_parent=False,
        ),
    ]

    result = solver.solve_column(column, signals)

    assert result.type_node == InferenceSignalType.ID.PRIMARY
    assert "UNIQUE" in result.reasons[0] and "PRIMARY" in result.reasons[1]


def test_complementary_signal_with_parent_trail() -> None:
    """Test that the solver will follow the weight trail and take complementary signals into account if parent has weight"""
    signals = [
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.UNKNOWN,
            reason="ID",
            confidence=InferenceSignalConfidence.HIGH,
            only_applies_to_parent=False,
        ),
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.UNIQUE,
            reason="UNIQUE",
            confidence=InferenceSignalConfidence.VERY_HIGH,
            only_applies_to_parent=True,
        ),
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.PRIMARY,
            reason="PRIMARY",
            confidence=InferenceSignalConfidence.VERY_HIGH,
            only_applies_to_parent=True,
        ),
    ]

    result = solver.solve_column(column, signals)

    assert result.type_node == InferenceSignalType.ID.PRIMARY
    assert "ID" in result.reasons[0] and "UNIQUE" in result.reasons[1] and "PRIMARY" in result.reasons[2]


def test_complementary_signals_without_parent_signal() -> None:
    """Test that the solver won't follow the weight trail and take complementary signals into account if parent has no weight"""
    signals = [
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.DIMENSION.CATEGORICAL,
            reason="CATEG_DIM",
            confidence=InferenceSignalConfidence.MEDIUM,
            only_applies_to_parent=False,
        ),
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.UNIQUE,
            reason="UNIQUE",
            confidence=InferenceSignalConfidence.VERY_HIGH,
            only_applies_to_parent=True,
        ),
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.PRIMARY,
            reason="PRIMARY",
            confidence=InferenceSignalConfidence.VERY_HIGH,
            only_applies_to_parent=True,
        ),
    ]

    result = solver.solve_column(column, signals)

    assert result.type_node == InferenceSignalType.DIMENSION.CATEGORICAL
    assert "CATEG_DIM" in result.reasons[0]


def test_contradicting_signals() -> None:
    """Test that the solver will return the deepest common ancestor if it finds conflicting signals."""
    signals = [
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.FOREIGN,
            reason="FOREIGN",
            confidence=InferenceSignalConfidence.HIGH,
            only_applies_to_parent=False,
        ),
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.PRIMARY,
            reason="PRIMARY",
            confidence=InferenceSignalConfidence.HIGH,
            only_applies_to_parent=False,
        ),
    ]

    result = solver.solve_column(column, signals)

    assert result.type_node == InferenceSignalType.ID.UNKNOWN


def test_stop_at_internal_node_if_trail_stops() -> None:
    """Test that if the signal trail stops at an internal node the solver will return that node instead of going deeper."""
    signals = [
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.UNKNOWN,
            reason="KEY",
            confidence=InferenceSignalConfidence.HIGH,
            only_applies_to_parent=False,
        ),
        InferenceSignal(
            column=column,
            type_node=InferenceSignalType.ID.UNIQUE,
            reason="UNIQUE",
            confidence=InferenceSignalConfidence.HIGH,
            only_applies_to_parent=False,
        ),
    ]

    result = solver.solve_column(column, signals)

    # should not progress further into the tree and assume it's PRIMARY
    assert result.type_node == InferenceSignalType.ID.UNIQUE
    assert "KEY" in result.reasons[0] and "UNIQUE" in result.reasons[1]



from __future__ import annotations

from dataclasses import dataclass

from metricflow.dataclass_serialization import SerializableDataclass


@dataclass(frozen=True)
class ElementReference(SerializableDataclass):
    """Used when we need to refer to a dimension, measure, identifier, but other attributes are unknown."""

    element_name: str


@dataclass(frozen=True)
class LinkableElementReference(ElementReference):
    """Used when we need to refer to a dimension or identifier, but other attributes are unknown."""

    pass


@dataclass(frozen=True)
class MeasureReference(ElementReference):
    """Used when we need to refer to a measure (separate from LinkableElementReference because measures aren't linkable"""

    pass


@dataclass(frozen=True)
class DimensionReference(LinkableElementReference):  # noqa: D
    pass

    @property
    def time_dimension_reference(self) -> TimeDimensionReference:  # noqa: D
        return TimeDimensionReference(element_name=self.element_name)


@dataclass(frozen=True)
class IdentifierReference(LinkableElementReference):  # noqa: D
    pass


@dataclass(frozen=True)
class CompositeSubIdentifierReference(ElementReference):  # noqa: D
    pass


@dataclass(frozen=True)
class TimeDimensionReference(DimensionReference):  # noqa: D
    pass

    def dimension_reference(self) -> DimensionReference:  # noqa: D
        return DimensionReference(element_name=self.element_name)


@dataclass(frozen=True)
class MetricReference(ElementReference):  # noqa: D
    pass



import traceback
from typing import List

from metricflow.errors.errors import ParsingException
from metricflow.instances import MetricModelReference
from metricflow.model.objects.metric import Metric, MetricType, CumulativeMetricWindow
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    FileContext,
    MetricContext,
    ModelValidationRule,
    ValidationIssueType,
    ValidationError,
    validate_safely,
)


class CumulativeMetricRule(ModelValidationRule):
    """Checks that cumulative sum metrics are configured properly"""

    @staticmethod
    @validate_safely(whats_being_done="checking that the params of metric are valid if it is a cumulative sum metric")
    def _validate_cumulative_sum_metric_params(metric: Metric) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        if metric.type == MetricType.CUMULATIVE:
            if metric.type_params.window and metric.type_params.grain_to_date:
                issues.append(
                    ValidationError(
                        context=MetricContext(
                            file_context=FileContext.from_metadata(metadata=metric.metadata),
                            metric=MetricModelReference(metric_name=metric.name),
                        ),
                        message="Both window and grain_to_date set for cumulative metric. Please set one or the other",
                    )
                )

            if metric.type_params.window:
                try:
                    _, _ = CumulativeMetricWindow.parse(metric.type_params.window.to_string())
                except ParsingException as e:
                    issues.append(
                        ValidationError(
                            context=MetricContext(
                                file_context=FileContext.from_metadata(metadata=metric.metadata),
                                metric=MetricModelReference(metric_name=metric.name),
                            ),
                            message="".join(traceback.format_exception_only(etype=type(e), value=e)),
                            extra_detail="".join(traceback.format_tb(e.__traceback__)),
                        )
                    )

        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring cumulative sum metrics are valid")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []

        for metric in model.metrics or []:
            issues += CumulativeMetricRule._validate_cumulative_sum_metric_params(metric=metric)

        return issues



from __future__ import annotations

import logging

from typing import Dict, List, Optional

from metricflow.configuration.config_handler import ConfigHandler
from metricflow.configuration.constants import CONFIG_DWH_SCHEMA
from metricflow.configuration.yaml_handler import YamlFileHandler
from metricflow.dataflow.sql_table import SqlTable
from metricflow.engine.metricflow_engine import (
    MetricFlowEngine,
    MetricFlowExplainResult,
    MetricFlowQueryRequest,
    MetricFlowQueryResult,
)
from metricflow.engine.models import Dimension, Materialization, Metric
from metricflow.engine.utils import build_user_configured_model_from_config, convert_to_datetime
from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.validations.validator_helpers import ModelValidationResults
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.optimizer.optimization_levels import SqlQueryOptimizationLevel
from metricflow.sql_clients.common_client import not_empty
from metricflow.sql_clients.sql_utils import make_sql_client_from_config

logger = logging.getLogger(__name__)


class MetricFlowClient:
    """MetricFlow Python client for running basic queries and other standard commands."""

    @staticmethod
    def from_config(config_file_path: Optional[str] = None) -> MetricFlowClient:
        """Builds a MetricFlowClient via config yaml file.

        If config_file_path is not passed, it will use the default config location.
        """
        if config_file_path is not None:
            handler = YamlFileHandler(yaml_file_path=config_file_path)
        else:
            handler = ConfigHandler()
        logger.debug(f"Constructing a MetricFlowClient with the config in {handler.yaml_file_path}")
        sql_client = make_sql_client_from_config(handler)
        user_configured_model = build_user_configured_model_from_config(handler)
        schema = not_empty(handler.get_value(CONFIG_DWH_SCHEMA), CONFIG_DWH_SCHEMA, handler.url)

        return MetricFlowClient(
            sql_client=sql_client,
            user_configured_model=user_configured_model,
            system_schema=schema,
        )

    def __init__(
        self,
        sql_client: SqlClient,
        user_configured_model: UserConfiguredModel,
        system_schema: str,
    ):
        """Initializer for MetricFlowClient.

        Args:
            sql_client: Client that is connected to your data warehouse.
            user_configured_model: Model containing all the information about your metric configs.
            system_schema: schema of where MF system tables are stored.
        """
        self.sql_client = sql_client
        self.user_configured_model = user_configured_model
        self.system_schema = system_schema
        self.semantic_model = SemanticModel(self.user_configured_model)
        self.engine = MetricFlowEngine(
            semantic_model=self.semantic_model,
            sql_client=self.sql_client,
            system_schema=self.system_schema,
        )

    def _create_mf_request(
        self,
        metrics: List[str],
        dimensions: List[str] = [],
        limit: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        where: Optional[str] = None,
        order: Optional[List[str]] = None,
        as_table: Optional[str] = None,
        sql_optimization_level: int = 4,
    ) -> MetricFlowQueryRequest:
        """Build MetricFlowQueryRequest given common query parameters."""
        parsed_optimization_level = SqlQueryOptimizationLevel(f"O{sql_optimization_level}")
        parsed_start_time = convert_to_datetime(start_time)
        parsed_end_time = convert_to_datetime(end_time)
        return MetricFlowQueryRequest.create_with_random_request_id(
            metric_names=metrics,
            group_by_names=dimensions,
            limit=limit,
            time_constraint_start=parsed_start_time,
            time_constraint_end=parsed_end_time,
            where_constraint=where,
            order_by_names=order,
            output_table=as_table,
            sql_optimization_level=parsed_optimization_level,
        )

    def query(
        self,
        metrics: List[str],
        dimensions: List[str] = [],
        limit: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        where: Optional[str] = None,
        order: Optional[List[str]] = None,
        as_table: Optional[str] = None,
        sql_optimization_level: int = 4,
    ) -> MetricFlowQueryResult:
        """Makes a query for a metric.

        Args:
            metrics: Names of the metrics to query.
            dimensions: Names of the dimensions and identifiers to query.
            limit: Limit the result to this many rows.
            start_time: Get data for the start of this time range.
            end_time: Get data for the end of this time range.
            where: A SQL string using group by names that can be used like a where clause on the output data.
            order: metric and group by names to order by. A "-" can be used to specify reverse order e.g. "-ds"
            as_table: If specified, output the result data to this table instead of a result dataframe.
            sql_optimization_level: The level of optimization for the generated SQL. Pass integer from 0-4.

        Returns:
            MetricFlowQueryResult that contains the result and context of the query.
        """
        mf_request = self._create_mf_request(
            metrics=metrics,
            dimensions=dimensions,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
            where=where,
            order=order,
            as_table=as_table,
            sql_optimization_level=sql_optimization_level,
        )
        return self.engine.query(mf_request=mf_request)

    def explain(
        self,
        metrics: List[str],
        dimensions: List[str] = [],
        limit: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        where: Optional[str] = None,
        order: Optional[List[str]] = None,
        as_table: Optional[str] = None,
        sql_optimization_level: int = 4,
    ) -> MetricFlowExplainResult:
        """Returns the plan for resolving a query.

        Args:
            metrics: Names of the metrics to query.
            dimensions: Names of the dimensions and identifiers to query.
            limit: Limit the result to this many rows.
            start_time: Get data for the start of this time range.
            end_time: Get data for the end of this time range.
            where: A SQL string using group by names that can be used like a where clause on the output data.
            order: metric and group by names to order by. A "-" can be used to specify reverse order e.g. "-ds"
            as_table: If specified, output the result data to this table instead of a result dataframe.
            sql_optimization_level: The level of optimization for the generated SQL. Pass integer from 0-4.

        Returns:
            MetricFlowExplainResult that contains the context of the query.
        """
        mf_request = self._create_mf_request(
            metrics=metrics,
            dimensions=dimensions,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
            where=where,
            order=order,
            as_table=as_table,
            sql_optimization_level=sql_optimization_level,
        )
        return self.engine.explain(mf_request=mf_request)

    def list_metrics(self) -> Dict[str, Metric]:
        """Retrieves a list of metric names.

        Returns:
            A dictionary with metric names as the key and the corresponding Metric object as the value.
        """
        return {m.name: m for m in self.engine.list_metrics()}

    def list_dimensions(self, metric_names: List[str]) -> List[Dimension]:
        """Retrieves a list of all common dimensions for metric_names.

        "simple" dimensions are the ones that people expect from a UI perspective. For example, if "ds" is a time
        dimension at a day granularity, this would not list "ds__week".

        Args:
            metric_names: Names of metrics to get common dimensions from.

        Returns:
            A list of Dimension objects containing metadata.
        """
        return self.engine.simple_dimensions_for_metrics(metric_names=metric_names)

    def list_materializations(self) -> List[Materialization]:
        """Retrieves a list of materialization names.

        Returns:
            A list of Materialization objects containing metadata.
        """
        return self.engine.list_materializations()

    def get_dimension_values(
        self, metric_name: str, dimension_name: str, start_time: Optional[str] = None, end_time: Optional[str] = None
    ) -> List[str]:
        """Retrieves a list of dimension values given a [metric_name, dimension_name].

        Args:
            metric_name: Name of metric that contains the group_by.
            dimension_name: Name of group_by to get values from.
            start_time: Get data for the start of this time range.
            end_time: Get data for the end of this time range.

        Returns:
            A list of dimension values as string.
        """
        parsed_start_time = convert_to_datetime(start_time)
        parsed_end_time = convert_to_datetime(end_time)
        return self.engine.get_dimension_values(
            metric_name=metric_name,
            get_group_by_values=dimension_name,
            time_constraint_start=parsed_start_time,
            time_constraint_end=parsed_end_time,
        )

    def materialize(
        self, materialization_name: str, start_time: Optional[str] = None, end_time: Optional[str] = None
    ) -> SqlTable:
        """Builds a table containing metrics and dimensions from a materialization definition.

        This can be very expensive if a large time range is provided.

        Args:
            materialization_name: Name of materialization
            start_time: Materialized for the start of this time range.
            end_time: Materialized for the end of this time range.

        Returns:
            SqlTable object of the materialized table.
        """
        parsed_start_time = convert_to_datetime(start_time)
        parsed_end_time = convert_to_datetime(end_time)
        return self.engine.materialize(
            materialization_name=materialization_name,
            time_constraint_start=parsed_start_time,
            time_constraint_end=parsed_end_time,
        )

    def drop_materialization(self, materialization_name: str) -> bool:
        """Drops the table associated with a materialization definition.

        Args:
            materialization_name: Name of materialization to drop.

        Returns:
            True if a table has been drop, False if table doesn't exist.
        """
        return self.engine.drop_materialization(materialization_name=materialization_name)

    def validate_configs(self) -> ModelValidationResults:
        """Validate a model according to configured rules.

        Returns:
            Tuple of validation issues with the model provided.
        """
        return ModelValidator().validate_model(self.user_configured_model).issues



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import List, Optional, Tuple, Sequence

from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlQueryPlanOptimizer
from metricflow.sql.sql_exprs import (
    SqlColumnReference,
    SqlExpressionTreeLineage,
    SqlExpressionNode,
    SqlColumnReplacements,
    SqlLogicalExpression,
    SqlLogicalOperator,
    SqlColumnAliasReferenceExpression,
)
from metricflow.sql.sql_plan import (
    SqlQueryPlanNode,
    SqlQueryPlanNodeVisitor,
    SqlSelectQueryFromClauseNode,
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlOrderByDescription,
    SqlJoinDescription,
    SqlSelectColumn,
)

logger = logging.getLogger(__name__)


@dataclass
class RewritableSqlClauses:
    """Stores clauses in a SQL query that should be rewritten when a node is reduced."""

    select_columns: List[SqlSelectColumn]
    wheres: List[SqlExpressionNode]
    group_bys: List[SqlSelectColumn]
    order_bys: List[SqlOrderByDescription]

    def rewrite(self, column_replacements: SqlColumnReplacements) -> None:
        """Rewrite all clauses using the given replacements."""
        self.select_columns = [
            SqlSelectColumn(expr=x.expr.rewrite(column_replacements), column_alias=x.column_alias)
            for x in self.select_columns
        ]
        self.wheres = [x.rewrite(column_replacements=column_replacements) for x in self.wheres]
        self.group_bys = [
            SqlSelectColumn(expr=x.expr.rewrite(column_replacements), column_alias=x.column_alias)
            for x in self.group_bys
        ]
        self.order_bys = [
            SqlOrderByDescription(
                expr=x.expr.rewrite(column_replacements=column_replacements),
                desc=x.desc,
            )
            for x in self.order_bys
        ]

    def combine_wheres(self, additional_where_clauses: List[SqlExpressionNode]) -> Optional[SqlExpressionNode]:
        """Combine the WHERE clauses in this with the additional clauses to form a single WHERE clause."""
        all_where_clauses = self.wheres + additional_where_clauses
        if len(all_where_clauses) == 1:
            return all_where_clauses[0]
        elif len(all_where_clauses) > 1:
            return SqlLogicalExpression(
                operator=SqlLogicalOperator.AND,
                args=tuple(all_where_clauses),
            )
        return None

    @property
    def contains_ambiguous_exprs(self) -> bool:
        """Returns true if any of the clauses have ambiguous expressions that will be difficult to re-write"""
        return any(
            [x.expr.lineage.contains_ambiguous_exprs for x in self.select_columns]
            + [x.lineage.contains_ambiguous_exprs for x in self.wheres]
            + [x.expr.lineage.contains_ambiguous_exprs for x in self.group_bys]
            + [x.expr.lineage.contains_ambiguous_exprs for x in self.order_bys]
        )


class SqlRewritingSubQueryReducerVisitor(SqlQueryPlanNodeVisitor[SqlQueryPlanNode]):
    """Visits the SQL query plan to simplify sub-queries. On each visit, return a simplified node.

    Unlike SqlSubQueryReducerVisitor, this will re-write expressions to realize more reductions.
    """

    def _reduce_parents(
        self,
        node: SqlSelectStatementNode,
    ) -> SqlSelectStatementNode:
        """Apply the reducing operation to the parent select statements."""
        return SqlSelectStatementNode(
            description=node.description,
            select_columns=node.select_columns,
            from_source=node.from_source.accept(self),
            from_source_alias=node.from_source_alias,
            joins_descs=tuple(
                SqlJoinDescription(
                    right_source=x.right_source.accept(self),
                    right_source_alias=x.right_source_alias,
                    on_condition=x.on_condition,
                    join_type=x.join_type,
                )
                for x in node.join_descs
            ),
            group_bys=node.group_bys,
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
        )

    @staticmethod
    def _statement_contains_difficult_expressions(node: SqlSelectStatementNode) -> bool:
        combined_lineage = SqlExpressionTreeLineage.combine(
            tuple(x.expr.lineage for x in node.select_columns)
            + ((node.where.lineage,) if node.where else ())
            + tuple(x.expr.lineage for x in node.group_bys)
            + tuple(x.expr.lineage for x in node.order_bys)
        )

        return combined_lineage.contains_string_exprs or combined_lineage.contains_column_alias_exprs

    @staticmethod
    def _select_columns_contain_string_expressions(select_columns: Tuple[SqlSelectColumn, ...]) -> bool:
        combined_lineage = SqlExpressionTreeLineage.combine(tuple(x.expr.lineage for x in select_columns))

        return len(combined_lineage.string_exprs) > 0

    @staticmethod
    def _select_columns_are_column_references(select_columns: Tuple[SqlSelectColumn, ...]) -> bool:
        for select_column in select_columns:
            if not select_column.expr.as_column_reference_expression:
                return False
        return True

    @staticmethod
    def _select_column_for_alias(column_alias: str, select_columns: Sequence[SqlSelectColumn]) -> SqlSelectColumn:
        for select_column in select_columns:
            if select_column.column_alias == column_alias:
                return select_column
        raise RuntimeError(f"Column alias '{column_alias}' not in SELECT columns: {select_columns}")

    @staticmethod
    def _is_simple_source(node: SqlSelectStatementNode) -> bool:
        """Returns true if the node is simple.

        Simple is defined as having no JOINs, WHERE, GROUP BYs, ORDER BYs, LIMIT and there are no strings in the column
        select. Strings are avoided so that the child node doesn't use the string expression in a group by or cause
        aliasing issues when used in the child query.

        e.g.

        Reducing

        SELECT
          a.col AS a_col
          , b.col AS b_col
        FROM (
           SELECT
             col AS col -- "col" is a string expression, not a column reference expression.
             , src0.join_key AS join_key
           FROM foo src0
        ) a
        JOIN b
        ON a.join_key = b.join_key

        to

        SELECT
          col AS a_col
          , b.col AS b_col
        FROM foo src0
        JOIN b
        ON a.join_key = b.join_key

        will throw an ambiguous column error on "col".
        """
        for select_column in node.select_columns:
            if select_column.expr.lineage.contains_string_exprs:
                return False
            if select_column.expr.lineage.contains_column_alias_exprs:
                return False
        return (
            len(node.parent_nodes) <= 1
            and len(node.group_bys) == 0
            and len(node.order_bys) == 0
            and not node.limit
            and not node.where
        )

    def _current_node_can_be_reduced(self, node: SqlSelectStatementNode) -> bool:  # noqa: D
        """Returns true if the given node can be reduced with the parent node.

        Reducing this node means eliminating the SELECT of this node and merging it with the parent SELECT. This
        checks for the cases where we are able to reduce.
        """

        # If this node has multiple parents (i.e. a join) that are complex, then this can't be collapsed.
        is_join = len(node.join_descs) > 0
        has_multiple_parent_nodes = len(node.parent_nodes) > 1
        if has_multiple_parent_nodes or is_join:
            return False

        assert len(node.parent_nodes) == 1
        parent_node = node.parent_nodes[0]

        # If the parent node is not a SELECT statement, then this can't be collapsed. e.g. with a table as a parent like
        # SELECT foo FROM bar
        if not parent_node.as_select_node:
            return False
        parent_select_node = parent_node.as_select_node

        # More conditions where we don't want to collapse. It's not impossible with these cases, but not reducing in
        # these cases for simplicity.

        # Re-writing string expressions / column alias expressions not yet supported, so don't reduce in those cases.
        if SqlRewritingSubQueryReducerVisitor._statement_contains_difficult_expressions(node):
            return False

        # Skip this case for simplicity of reasoning.
        if len(node.order_bys) > 0 and len(parent_select_node.order_bys) > 0:
            return False

        # Skip this case for simplicity of reasoning.
        if len(parent_select_node.group_bys) > 0 and len(node.group_bys) > 0:
            return False

        # TODO: Check for the following case:
        # SELECT
        #   bookings
        #   , 2 * bookings AS twice_bookings
        # FROM (
        #   SELECT,
        #     SUM(bookings) AS bookings
        #     , fct_bookings_src.is_instant
        #   FROM (
        #     SELECT * FROM demo.fct_bookings
        #   ) fct_bookings_src
        #   GROUP BY fct_bookings_src.is_instant
        # ) src
        #
        # If this is reduced, then the GROUP BY will refer to an unused column.

        # Don't reduce if the ORDER BYs aren't column reference expressions for simplicity.
        for order_by in node.order_bys:
            order_by_column_reference_expression = order_by.expr.as_column_reference_expression
            if not order_by_column_reference_expression:
                return False
            # Also for simplicity, the ORDER BY must match one of the SELECT expressions.
            if not SqlRewritingSubQueryReducerVisitor._find_matching_select_column(
                col_ref=order_by_column_reference_expression.col_ref,
                select_columns=node.select_columns,
            ):
                return False

        # If the parent has a GROUP BY and this has a WHERE, avoid reducing as the WHERE could reference an
        # aggregation expression.
        if len(parent_select_node.group_bys) > 0 and node.where:
            return False

        # If the parent has a GROUP BY, the case where it's easiest to merge this with the parent is if all select
        # columns are column references.
        if len(
            parent_select_node.group_bys
        ) > 0 and not SqlRewritingSubQueryReducerVisitor._select_columns_are_column_references(node.select_columns):
            return False

        # If the parent select node contains string columns, and this has a GROUP BY, don't reduce as string columns
        # can have special meanings in a GROUP BY. For example,
        #
        # SELECT
        #   a.col1 AS col1
        #   , a.col2 AS col2
        # FROM (
        #   SELECT
        #     b.col1 AS col1
        #     , 1 AS col2
        #   FROM foo b
        # ) a
        # GROUP BY a.col2
        #
        # would be rewritten in the current algorithm to
        #
        # SELECT
        #   b.col1 AS col1
        #   , 1 AS col2
        # FROM foo b
        # GROUP by 1
        #
        # However, the GROUP BY would be wrong as the "1" string expression would be interpreted as selecting the 1st
        # item in the SELECT to group by.

        if len(node.group_bys) > 0 and SqlRewritingSubQueryReducerVisitor._select_columns_contain_string_expressions(
            select_columns=parent_select_node.select_columns,
        ):
            return False

        return True

    @staticmethod
    def _find_matching_table_alias(node: SqlSelectStatementNode, column_alias: str) -> Optional[str]:
        for select_column in node.select_columns:
            if select_column.column_alias == column_alias:
                column_reference_expr = select_column.expr.as_column_reference_expression
                if column_reference_expr:
                    return column_reference_expr.col_ref.table_alias
        return None

    @staticmethod
    def _get_column_replacements(parent_node: SqlSelectStatementNode, parent_node_alias: str) -> SqlColumnReplacements:
        column_replacements = {}
        for select_column in parent_node.select_columns:
            column_reference = SqlColumnReference(
                table_alias=parent_node_alias,
                column_name=select_column.column_alias,
            )
            column_replacements[column_reference] = select_column.expr

        return SqlColumnReplacements(column_replacements)

    @staticmethod
    def _rewrite_select_columns(
        old_select_columns: Tuple[SqlSelectColumn, ...], column_replacements: SqlColumnReplacements
    ) -> Tuple[SqlSelectColumn, ...]:
        return tuple(
            SqlSelectColumn(expr=x.expr.rewrite(column_replacements), column_alias=x.column_alias)
            for x in old_select_columns
        )

    @staticmethod
    def _rewrite_where(
        column_replacements: SqlColumnReplacements,
        node_where: Optional[SqlExpressionNode] = None,
        parent_node_where: Optional[SqlExpressionNode] = None,
    ) -> Optional[SqlExpressionNode]:
        if node_where is None and parent_node_where is None:
            return None
        elif node_where and parent_node_where is None:
            return node_where.rewrite(column_replacements)
        elif node_where is None and parent_node_where:
            return parent_node_where

        # For type checking. The above conditionals should ensure the below.
        assert node_where
        assert parent_node_where
        return SqlLogicalExpression(operator=SqlLogicalOperator.AND, args=(node_where, parent_node_where))

    @staticmethod
    def _find_matching_select_column(
        col_ref: SqlColumnReference, select_columns: Sequence[SqlSelectColumn]
    ) -> Optional[SqlSelectColumn]:
        for select_column in select_columns:
            column_reference_expression = select_column.expr.as_column_reference_expression
            if column_reference_expression and column_reference_expression.col_ref == col_ref:
                return select_column
        return None

    @staticmethod
    def _rewrite_node_with_join(node: SqlSelectStatementNode) -> SqlSelectStatementNode:
        """Reduces nodes with joins if the join source is simple to reduce.

        Converts this:

        SELECT
          SUM(bookings_src.bookings) AS bookings
          listings_src.country_latest AS listing__country_latest
          bookings_src.ds AS ds
        FROM (
          SELECT
            fct_bookings_src.booking AS bookings
            , fct_bookings_src.ds AS ds
            , fct_bookings_src.listing_id AS listing
          FROM demo.fct_bookings fct_bookings_src
        ) bookings_src
        JOIN (
          SELECT
            dim_listings_src.country_latest AS country_latest
            , listing_id AS listing
          FROM demo.dim_listings dim_listings_src
        ) listings_src
        ON bookings_src.listing = listings_src.listing
        GROUP BY bookings_src.ds

        to:

        SELECT
          SUM(bookings_src.booking) AS bookings
          dim_listings_src.country_latest AS listing__country_latest
          bookings_src.ds AS ds
        FROM demo.fct_bookings fct_bookings_src
        JOIN demo.dim_listings dim_listings_src
        ON bookings_src.listing_id = dim_listings_src.listing_id
        GROUP BY bookings_src.ds
        """

        # Check that there aren't any duplicates in source aliases, or else there would be a collision when reduced.
        # This check is conservative as it checks for duplicates in this node and parent nodes, but depending on
        # on which sources get reduced, there may not be a collision.
        from_source = node.from_source
        from_source_select = from_source.as_select_node
        from_source_alias = node.from_source_alias

        all_source_aliases = [from_source_alias]
        source_alias_set = {from_source_alias}

        if from_source_select:
            all_source_aliases.append(from_source_select.from_source_alias)
            source_alias_set.add(from_source_select.from_source_alias)

        for join_desc in node.join_descs:
            all_source_aliases.append(join_desc.right_source_alias)
            source_alias_set.add(join_desc.right_source_alias)
            joined_node_select = join_desc.right_source.as_select_node
            if joined_node_select:
                all_source_aliases.append(joined_node_select.from_source_alias)
                source_alias_set.add(joined_node_select.from_source_alias)

        # Some duplicate aliases, so safer to not reduce.
        if len(all_source_aliases) != len(source_alias_set):
            return node

        # See if any of the joined sources can be reduced.
        new_join_descs = []

        clauses_to_rewrite = RewritableSqlClauses(
            select_columns=list(node.select_columns),
            wheres=[node.where] if node.where else [],
            group_bys=list(node.group_bys),
            order_bys=list(node.order_bys),
        )

        # Can't re-write ambiguous expressions.
        if clauses_to_rewrite.contains_ambiguous_exprs:
            return node

        additional_where_clauses = []
        column_replacements_from_all_joins = []

        for join_desc in node.join_descs:
            join_select_node = join_desc.right_source.as_select_node

            # Verifying that it's simple makes it easier to reason about the logic.
            if not join_select_node or not SqlRewritingSubQueryReducerVisitor._is_simple_source(join_select_node):
                new_join_descs.append(join_desc)
                continue

            column_replacements = SqlRewritingSubQueryReducerVisitor._get_column_replacements(
                parent_node=join_select_node,
                parent_node_alias=join_desc.right_source_alias,
            )
            column_replacements_from_all_joins.append(column_replacements)

            new_join_descs.append(
                SqlJoinDescription(
                    right_source=join_select_node.from_source,
                    right_source_alias=join_select_node.from_source_alias,
                    on_condition=join_desc.on_condition.rewrite(column_replacements),
                    join_type=join_desc.join_type,
                )
            )

            if join_select_node.where:
                additional_where_clauses.append(join_select_node.where)

            clauses_to_rewrite.rewrite(column_replacements)

        # While the above re-wrote each individual join when that join was reduced, the ON condition could reference
        # columns from other joins, so the column replacements have to be applied to all join ON conditions.
        for column_replacements in column_replacements_from_all_joins:
            new_join_descs = [
                SqlJoinDescription(
                    right_source=x.right_source,
                    right_source_alias=x.right_source_alias,
                    on_condition=x.on_condition.rewrite(column_replacements=column_replacements),
                    join_type=x.join_type,
                )
                for x in new_join_descs
            ]

        from_source_is_simple = (
            SqlRewritingSubQueryReducerVisitor._is_simple_source(from_source_select) if from_source_select else False
        )
        if from_source_select and from_source_is_simple:
            column_replacements = SqlRewritingSubQueryReducerVisitor._get_column_replacements(
                parent_node=from_source_select,
                parent_node_alias=node.from_source_alias,
            )

            if from_source_select.where:
                additional_where_clauses.append(from_source_select.where)

            clauses_to_rewrite.rewrite(column_replacements=column_replacements)
            # This was already checked in _is_simple_source().
            assert len(from_source_select.parent_nodes) == 1
            from_source = from_source_select.from_source
            from_source_alias = from_source_select.from_source_alias

            new_join_descs = [
                SqlJoinDescription(
                    right_source=x.right_source,
                    right_source_alias=x.right_source_alias,
                    on_condition=x.on_condition.rewrite(column_replacements=column_replacements),
                    join_type=x.join_type,
                )
                for x in new_join_descs
            ]

        return SqlSelectStatementNode(
            description=node.description,
            select_columns=tuple(clauses_to_rewrite.select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=tuple(new_join_descs),
            group_bys=tuple(clauses_to_rewrite.group_bys),
            order_bys=tuple(clauses_to_rewrite.order_bys),
            where=clauses_to_rewrite.combine_wheres(additional_where_clauses),
            limit=node.limit,
        )

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlQueryPlanNode:  # noqa: D
        node_with_reduced_parents = self._reduce_parents(node)

        if len(node_with_reduced_parents.parent_nodes) > 1:
            return SqlRewritingSubQueryReducerVisitor._rewrite_node_with_join(node_with_reduced_parents)

        if not self._current_node_can_be_reduced(node_with_reduced_parents):
            return node_with_reduced_parents

        # Note that just because this node can't be reduced doesn't mean the parent node can't. In this example, the
        # outer query can't be reduced because there is a join, but sub-query d can be.
        #
        # SELECT a.bookings, b.listing__country
        # FROM (
        #   SELECT d.bookings, d.listing_id
        #   FROM (
        #     SELECT e.bookings, e.listing_id
        #     FROM fct_bookings
        #   ) d
        #   LIMIT 10
        # ) b
        # JOIN dim_listings c
        # ON a.listing_id = b.listing_id

        assert len(node_with_reduced_parents.parent_nodes) == 1
        parent_node = node_with_reduced_parents.parent_nodes[0]
        parent_select_node = parent_node.as_select_node
        assert parent_select_node

        # At this point, the query should look similar to
        #
        #     SELECT b.foo, b.baz
        #     FROM (
        #       SELECT 1 AS foo, a.baz FROM bar a
        #       LIMIT 10
        #     ) b
        #     GROUP BY b.foo
        #     ORDER by b.baz
        #     LIMIT 1

        # The ORDER BY in the parent doesn't matter since the order by in this node will "overwrite" the order in the
        # parent as long as the parent has no limits.
        column_replacements = SqlRewritingSubQueryReducerVisitor._get_column_replacements(
            parent_node=parent_select_node,
            parent_node_alias=node.from_source_alias,
        )
        new_order_bys: List[SqlOrderByDescription] = []
        # Handle ORDER BY differently to avoid this hazard with expression rewriting:
        #
        # SELECT a.bookings AS bookings
        # FROM (
        #   SELECT SUM(b.bookings) AS bookings
        #   FROM fct_bookings b
        # ) a
        # ORDER BY a.bookings
        #
        # ->
        #
        # SELECT SUM(b.bookings) AS bookings
        # FROM fct_bookings b
        # ORDER BY SUM(b.bookings) -- Throws an error in some engines.
        if node_with_reduced_parents.order_bys:
            for order_by_item in node_with_reduced_parents.order_bys:
                order_by_item_expr = order_by_item.expr.as_column_reference_expression
                assert order_by_item_expr

                matching_select_column = SqlRewritingSubQueryReducerVisitor._find_matching_select_column(
                    col_ref=order_by_item_expr.col_ref,
                    select_columns=node_with_reduced_parents.select_columns,
                )
                # This must be the case because of _should_reduce()
                assert matching_select_column

                new_order_bys.append(
                    SqlOrderByDescription(
                        expr=SqlColumnAliasReferenceExpression(column_alias=matching_select_column.column_alias),
                        desc=order_by_item.desc,
                    )
                )

        # The limit should be the min of this SELECT limit and the parent SELECT limit.
        new_limit: Optional[int] = node_with_reduced_parents.limit
        if new_limit is None:
            new_limit = parent_select_node.limit
        elif parent_select_node.limit is not None:
            new_limit = min(new_limit, parent_select_node.limit)

        new_group_bys: Tuple[SqlSelectColumn, ...] = ()
        if node.group_bys and parent_select_node.group_bys:
            raise RuntimeError(
                "Attempting to reduce sub-queries when this and the parent have GROUP BYs. This should have been "
                "prevent by _should_reduce()"
            )
        elif node.group_bys:
            new_group_bys = SqlRewritingSubQueryReducerVisitor._rewrite_select_columns(
                old_select_columns=node.group_bys, column_replacements=column_replacements
            )
        elif parent_select_node.group_bys:
            new_group_bys = parent_select_node.group_bys

        return SqlSelectStatementNode(
            description="\n".join([parent_select_node.description, node_with_reduced_parents.description]),
            select_columns=SqlRewritingSubQueryReducerVisitor._rewrite_select_columns(
                old_select_columns=node.select_columns, column_replacements=column_replacements
            ),
            from_source=parent_select_node.from_source,
            from_source_alias=parent_select_node.from_source_alias,
            joins_descs=parent_select_node.join_descs,
            group_bys=new_group_bys,
            order_bys=tuple(new_order_bys),
            where=SqlRewritingSubQueryReducerVisitor._rewrite_where(
                column_replacements=column_replacements,
                node_where=node.where,
                parent_node_where=parent_select_node.where,
            ),
            limit=new_limit,
        )

    def visit_table_from_clause_node(self, node: SqlTableFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        return node


class SqlGroupByRewritingVisitor(SqlQueryPlanNodeVisitor[SqlQueryPlanNode]):
    """Re-writes the GROUP BY to use a SqlColumnAliasReferenceExpression."""

    @staticmethod
    def _find_matching_select(
        expr: SqlExpressionNode, select_columns: Sequence[SqlSelectColumn]
    ) -> Optional[SqlSelectColumn]:
        """Given an expression, find the SELECT column that has the same expression."""
        for select_column in select_columns:
            if select_column.expr == expr:
                return select_column
        return None

    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> SqlQueryPlanNode:  # noqa: D
        new_group_bys = []
        for group_by in node.group_bys:
            matching_select_column = SqlGroupByRewritingVisitor._find_matching_select(
                group_by.expr, node.select_columns
            )
            if matching_select_column:
                new_group_bys.append(
                    SqlSelectColumn(
                        expr=SqlColumnAliasReferenceExpression(column_alias=matching_select_column.column_alias),
                        column_alias=matching_select_column.column_alias,
                    )
                )
            else:
                logger.error(f"Did not find matching select for {group_by} in {node}")
                new_group_bys.append(group_by)

        return SqlSelectStatementNode(
            description=node.description,
            select_columns=node.select_columns,
            from_source=node.from_source.accept(self),
            from_source_alias=node.from_source_alias,
            joins_descs=tuple(
                SqlJoinDescription(
                    right_source=x.right_source.accept(self),
                    right_source_alias=x.right_source_alias,
                    on_condition=x.on_condition,
                    join_type=x.join_type,
                )
                for x in node.join_descs
            ),
            group_bys=tuple(new_group_bys),
            order_bys=node.order_bys,
            where=node.where,
            limit=node.limit,
        )

    def visit_table_from_clause_node(self, node: SqlTableFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        return node

    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> SqlQueryPlanNode:  # noqa: D
        return node


class SqlRewritingSubQueryReducer(SqlQueryPlanOptimizer):
    """Simplify queries by eliminating sub-queries when possible by rewriting expressions.

     Expressions in the SELECT, GROUP BY, and WHERE are can be rewritten.

    e.g. from

    SELECT b.col0 AS foo
    FROM (
      SELECT SUM(a.col0) AS bar
      FROM table0 a
    ) b
    GROUP BY b.col0

    to

    SELECT SUM(a.col0) AS foo
    FROM table0 a
    GROUP BY foo
    """

    def __init__(self, use_column_alias_in_group_bys: bool = False) -> None:  # noqa: D
        self._use_column_alias_in_group_bys = use_column_alias_in_group_bys

    def optimize(self, node: SqlQueryPlanNode) -> SqlQueryPlanNode:  # noqa: D
        result = node.accept(SqlRewritingSubQueryReducerVisitor())
        if self._use_column_alias_in_group_bys:
            return result.accept(SqlGroupByRewritingVisitor())
        return result



import pathlib
from typing import Any, List, Optional, Set

import dateutil.parser
import pandas as pd
from sqlalchemy.engine import make_url

from metricflow.configuration.constants import (
    CONFIG_DWH_CREDS_PATH,
    CONFIG_DWH_DB,
    CONFIG_DWH_DIALECT,
    CONFIG_DWH_HOST,
    CONFIG_DWH_PASSWORD,
    CONFIG_DWH_PORT,
    CONFIG_DWH_PROJECT_ID,
    CONFIG_DWH_USER,
    CONFIG_DWH_WAREHOUSE,
    CONFIG_DWH_ACCESS_TOKEN,
    CONFIG_DWH_HTTP_PATH,
)
from metricflow.configuration.yaml_handler import YamlFileHandler
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql_clients.big_query import BigQuerySqlClient
from metricflow.sql_clients.common_client import SqlDialect, not_empty
from metricflow.sql_clients.duckdb import DuckDbSqlClient
from metricflow.sql_clients.postgres import PostgresSqlClient
from metricflow.sql_clients.redshift import RedshiftSqlClient
from metricflow.sql_clients.snowflake import SnowflakeSqlClient
from metricflow.sql_clients.databricks import DatabricksSqlClient


def make_df(  # type: ignore [misc]
    sql_client: SqlClient, columns: List[str], data: Any, time_columns: Optional[Set[str]] = None
) -> pd.DataFrame:
    """Helper to make a dataframe, converting the time columns to appropriate types."""
    time_columns = time_columns or set()
    # Should only be used in testing, so sql_client should be set.
    assert sql_client

    if sql_client.sql_engine_attributes.timestamp_type_supported:
        new_rows = []
        for row in data:
            new_row = []
            # Change the type of the column if it's in time_columns
            for i, column in enumerate(columns):
                if column in time_columns and row[i] is not None:
                    # ts_suffix = " 00:00:00" if ":" not in row[i] else ""
                    # ts_input = row[i] + ts_suffix
                    new_row.append(dateutil.parser.parse(row[i]))

                else:
                    new_row.append(row[i])
            new_rows.append(new_row)
        data = new_rows

    return pd.DataFrame(
        columns=columns,
        data=data,
    )


def make_sql_client(url: str, password: str) -> SqlClient:
    """Build SQL client based on env configs. Used only in tests."""
    dialect_protocol = make_url(url.split(";")[0]).drivername.split("+")
    dialect = SqlDialect(dialect_protocol[0])
    if len(dialect_protocol) > 2:
        raise ValueError(f"Invalid # of +'s in {url}")

    if dialect == SqlDialect.REDSHIFT:
        return RedshiftSqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.SNOWFLAKE:
        return SnowflakeSqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.BIGQUERY:
        return BigQuerySqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.POSTGRESQL:
        return PostgresSqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.DUCKDB:
        return DuckDbSqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.DATABRICKS:
        return DatabricksSqlClient.from_connection_details(url, password)
    else:
        raise ValueError(f"Unknown dialect: `{dialect}` in URL {url}")


def make_sql_client_from_config(handler: YamlFileHandler) -> SqlClient:
    """Construct a SqlClient given a yaml file config."""

    url = handler.url
    dialect = not_empty(handler.get_value(CONFIG_DWH_DIALECT), CONFIG_DWH_DIALECT, url).lower()
    if dialect == SqlDialect.BIGQUERY.value:
        path_to_creds = handler.get_value(CONFIG_DWH_CREDS_PATH)
        project_id = handler.get_value(CONFIG_DWH_PROJECT_ID) or ""
        creds = None

        if not any([path_to_creds, project_id]):
            raise ValueError(f"One of `{CONFIG_DWH_CREDS_PATH}` or `{CONFIG_DWH_PROJECT_ID}` should be filled.")

        if path_to_creds:
            # Load json credential (Only for service accounts auth)
            if not pathlib.Path(path_to_creds).exists:
                raise ValueError(f"`{path_to_creds}` does not exist.")
            with open(path_to_creds, "r") as cred_file:
                creds = cred_file.read()
        return BigQuerySqlClient(project_id=project_id, password=creds)
    elif dialect == SqlDialect.SNOWFLAKE.value:
        host = not_empty(handler.get_value(CONFIG_DWH_HOST), CONFIG_DWH_HOST, url)
        user = not_empty(handler.get_value(CONFIG_DWH_USER), CONFIG_DWH_USER, url)
        password = not_empty(handler.get_value(CONFIG_DWH_PASSWORD), CONFIG_DWH_PASSWORD, url)
        database = not_empty(handler.get_value(CONFIG_DWH_DB), CONFIG_DWH_DB, url)
        warehouse = not_empty(handler.get_value(CONFIG_DWH_WAREHOUSE), CONFIG_DWH_WAREHOUSE, url)
        return SnowflakeSqlClient(
            host=host,
            username=user,
            password=password,
            database=database,
            url_query_params={"warehouse": warehouse},
            client_session_keep_alive=False,
        )
    elif dialect == SqlDialect.REDSHIFT.value:
        host = not_empty(handler.get_value(CONFIG_DWH_HOST), CONFIG_DWH_HOST, url)
        port = int(not_empty(handler.get_value(CONFIG_DWH_PORT), CONFIG_DWH_PORT, url))
        user = not_empty(handler.get_value(CONFIG_DWH_USER), CONFIG_DWH_USER, url)
        password = not_empty(handler.get_value(CONFIG_DWH_PASSWORD), CONFIG_DWH_PASSWORD, url)
        database = not_empty(handler.get_value(CONFIG_DWH_DB), CONFIG_DWH_DB, url)
        return RedshiftSqlClient(
            host=host,
            port=port,
            username=user,
            password=password,
            database=database,
        )
    elif dialect == SqlDialect.DUCKDB.value:
        database = not_empty(handler.get_value(CONFIG_DWH_DB), CONFIG_DWH_DB, url)
        return DuckDbSqlClient(file_path=database)
    elif dialect == SqlDialect.POSTGRESQL.value:
        host = not_empty(handler.get_value(CONFIG_DWH_HOST), CONFIG_DWH_HOST, url)
        port = int(not_empty(handler.get_value(CONFIG_DWH_PORT), CONFIG_DWH_PORT, url))
        user = not_empty(handler.get_value(CONFIG_DWH_USER), CONFIG_DWH_USER, url)
        password = not_empty(handler.get_value(CONFIG_DWH_PASSWORD), CONFIG_DWH_PASSWORD, url)
        database = not_empty(handler.get_value(CONFIG_DWH_DB), CONFIG_DWH_DB, url)
        return PostgresSqlClient(
            host=host,
            port=port,
            username=user,
            password=password,
            database=database,
        )
    elif dialect == SqlDialect.DATABRICKS.value:
        host = not_empty(handler.get_value(CONFIG_DWH_HOST), CONFIG_DWH_HOST, url)
        access_token = not_empty(handler.get_value(CONFIG_DWH_ACCESS_TOKEN), CONFIG_DWH_ACCESS_TOKEN, url)
        http_path = not_empty(handler.get_value(CONFIG_DWH_HTTP_PATH), CONFIG_DWH_HTTP_PATH, url)
        return DatabricksSqlClient(host=host, access_token=access_token, http_path=http_path)
    else:
        supported_dialects = [x.value for x in SqlDialect]
        raise ValueError(f"Invalid dialect '{dialect}', must be one of {supported_dialects} in {url}")



import logging
from collections import defaultdict
from dataclasses import dataclass
from datetime import date
from typing import List, MutableSet, Tuple, Sequence, DefaultDict

import more_itertools
from metricflow.instances import DataSourceElementReference, DataSourceReference

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.identifier import Identifier, IdentifierType, CompositeSubIdentifier
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    DataSourceContext,
    DataSourceElementContext,
    DataSourceElementType,
    FileContext,
    ModelValidationRule,
    ValidationIssue,
    ValidationError,
    ValidationIssueType,
    validate_safely,
    ValidationWarning,
)
from metricflow.model.validations.validator_helpers import ValidationFutureError
from metricflow.references import IdentifierReference

logger = logging.getLogger(__name__)


class IdentifierConfigRule(ModelValidationRule):
    """Checks that data source identifiers are valid"""

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring identifiers are valid")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues = []
        for data_source in model.data_sources:
            issues += IdentifierConfigRule._validate_data_source_identifiers(data_source=data_source)
        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking that the data source's identifiers are valid")
    def _validate_data_source_identifiers(data_source: DataSource) -> List[ValidationIssueType]:
        """Checks validity of composite identifiers"""
        issues: List[ValidationIssueType] = []
        for ident in data_source.identifiers:
            if ident.identifiers:
                context = DataSourceElementContext(
                    file_context=FileContext.from_metadata(metadata=data_source.metadata),
                    data_source_element=DataSourceElementReference(
                        data_source_name=data_source.name, element_name=ident.name
                    ),
                    element_type=DataSourceElementType.IDENTIFIER,
                )

                for sub_id in ident.identifiers:
                    if sub_id.ref and (sub_id.name or sub_id.expr):
                        logger.warning(f"Identifier with error is: {ident}")
                        issues.append(
                            ValidationError(
                                context=context,
                                message=f"Both ref and name/expr set in sub identifier of identifier "
                                f"({ident.name}), please set one",
                            )
                        )
                    elif sub_id.ref is not None and sub_id.ref not in [i.name for i in data_source.identifiers]:
                        issues.append(
                            ValidationError(
                                context=context,
                                message=f"Identifier ref must reference an existing identifier by name. "
                                f"No identifier in this data source has name: {sub_id.ref}",
                            )
                        )
                    elif not sub_id.ref and not sub_id.name:
                        issues.append(
                            ValidationError(
                                context=context,
                                message=f"Must provide either name or ref for sub identifier of identifier "
                                f"with name: {ident.reference.element_name}",
                            )
                        )

                    if sub_id.name:
                        for i in data_source.identifiers:
                            if i.name == sub_id.name and i.expr != sub_id.expr:
                                issues.append(
                                    ValidationError(
                                        context=context,
                                        message=f"If sub identifier has same name ({sub_id.name}) "
                                        f"as an existing Identifier they must have the same expr",
                                    )
                                )
                                break

        return issues


class OnePrimaryIdentifierPerDataSourceRule(ModelValidationRule):
    """Ensures that each data source has only one primary identifier"""

    @staticmethod
    @validate_safely(whats_being_done="checking data source has only one primary identifier")
    def _only_one_primary_identifier(data_source: DataSource) -> List[ValidationIssue]:
        primary_identifier_names: MutableSet[str] = set()
        for identifier in data_source.identifiers or []:
            if identifier.type == IdentifierType.PRIMARY:
                primary_identifier_names.add(identifier.reference.element_name)

        if len(primary_identifier_names) > 1:
            return [
                ValidationFutureError(
                    context=DataSourceContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source=DataSourceReference(data_source_name=data_source.name),
                    ),
                    message=f"Data sources can have only one primary identifier. The data source"
                    f" `{data_source.name}` has {len(primary_identifier_names)}: {', '.join(primary_identifier_names)}",
                    error_date=date(2022, 1, 12),  # Wed January 12th 2022
                )
            ]
        return []

    @staticmethod
    @validate_safely(
        whats_being_done="running model validation ensuring each data source has only one primary identifier"
    )
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssue]:  # noqa: D
        issues = []

        for data_source in model.data_sources:
            issues += OnePrimaryIdentifierPerDataSourceRule._only_one_primary_identifier(data_source=data_source)

        return issues


@dataclass(frozen=True)
class SubIdentifierContext:
    """Organizes the context behind identifiers and their sub-identifiers."""

    data_source: DataSource
    identifier_reference: IdentifierReference
    sub_identifier_names: Tuple[str, ...]


class IdentifierConsistencyRule(ModelValidationRule):
    """Checks identifiers with the same name are defined with the same set of sub-identifiers in all data sources"""

    @staticmethod
    def _get_sub_identifier_names(identifier: Identifier) -> Sequence[str]:
        sub_identifier_names = []
        sub_identifier: CompositeSubIdentifier
        for sub_identifier in identifier.identifiers or []:
            if sub_identifier.name:
                sub_identifier_names.append(sub_identifier.name)
            elif sub_identifier.ref:
                sub_identifier_names.append(sub_identifier.ref)
        return sub_identifier_names

    @staticmethod
    def _get_sub_identifier_context(data_source: DataSource) -> Sequence[SubIdentifierContext]:
        contexts = []
        for identifier in data_source.identifiers or []:
            contexts.append(
                SubIdentifierContext(
                    data_source=data_source,
                    identifier_reference=identifier.reference,
                    sub_identifier_names=tuple(IdentifierConsistencyRule._get_sub_identifier_names(identifier)),
                )
            )
        return contexts

    @staticmethod
    @validate_safely(whats_being_done="running model validation to ensure identifiers have consistent sub-identifiers")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []
        # build collection of sub-identifier contexts, keyed by identifier name
        identifier_to_sub_identifier_contexts: DefaultDict[str, List[SubIdentifierContext]] = defaultdict(list)
        all_contexts: List[SubIdentifierContext] = list(
            more_itertools.flatten(
                [
                    IdentifierConsistencyRule._get_sub_identifier_context(data_source)
                    for data_source in model.data_sources
                ]
            )
        )
        for context in all_contexts:
            identifier_to_sub_identifier_contexts[context.identifier_reference.element_name].append(context)

        # Filter out anything that has fewer than 2 distinct sub-identifier sets
        invalid_sub_identifier_configurations = dict(
            filter(
                lambda item: len(set([context.sub_identifier_names for context in item[1]])) >= 2,
                identifier_to_sub_identifier_contexts.items(),
            )
        )

        # convert each invalid identifier configuration into a validation warning
        for identifier_name, sub_identifier_contexts in invalid_sub_identifier_configurations.items():
            data_source = sub_identifier_contexts[0].data_source
            issues.append(
                ValidationWarning(
                    context=DataSourceElementContext(
                        file_context=FileContext.from_metadata(metadata=data_source.metadata),
                        data_source_element=DataSourceElementReference(
                            data_source_name=data_source.name, element_name=identifier_name
                        ),
                        element_type=DataSourceElementType.IDENTIFIER,
                    ),
                    message=(
                        f"Identifier '{identifier_name}' does not have consistent sub-identifiers "
                        f"throughout the model: {list(sorted(sub_identifier_contexts, key=lambda x: x.sub_identifier_names))}"
                    ),
                )
            )

        return issues



from __future__ import annotations

import click
import pytest

from click.testing import CliRunner, Result
from typing import Sequence

from metricflow.cli.cli_context import CLIContext
from metricflow.engine.metricflow_engine import MetricFlowEngine
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.test_utils import as_datetime
from metricflow.test.time.configurable_time_source import ConfigurableTimeSource


@pytest.fixture
def cli_context(  # noqa: D
    sql_client: SqlClient,
    simple_user_configured_model: UserConfiguredModel,
    time_spine_source: TimeSpineSource,
    mf_test_session_state: MetricFlowTestSessionState,
    create_simple_model_tables: bool,
) -> CLIContext:
    semantic_model = SemanticModel(simple_user_configured_model)
    mf_engine = MetricFlowEngine(
        semantic_model=semantic_model,
        sql_client=sql_client,
        column_association_resolver=DefaultColumnAssociationResolver(semantic_model=semantic_model),
        time_source=ConfigurableTimeSource(as_datetime("2020-01-01")),
        time_spine_source=time_spine_source,
        system_schema=mf_test_session_state.mf_system_schema,
    )
    context = CLIContext()
    context._mf = mf_engine
    context._sql_client = sql_client
    context._user_configured_model = simple_user_configured_model
    context._semantic_model = semantic_model
    context._mf_system_schema = mf_test_session_state.mf_system_schema
    return context


class MetricFlowCliRunner(CliRunner):
    """Custom CliRunner class to handle passing context."""

    def __init__(self, cli_context: CLIContext) -> None:  # noqa: D
        self.cli_context = cli_context
        super().__init__()

    def run(self, cli: click.BaseCommand, args: Sequence[str] = None) -> Result:  # noqa: D
        return super().invoke(cli, args, obj=self.cli_context)


@pytest.fixture
def cli_runner(cli_context: CLIContext) -> MetricFlowCliRunner:  # noqa: D
    return MetricFlowCliRunner(cli_context=cli_context)



from __future__ import annotations

import logging
import os
from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Dict, Sequence

import pytest

from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataflow.dataflow_plan import ReadSqlSourceNode, BaseOutput
from metricflow.dataset.convert_data_source import DataSourceToDataSetConverter
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.dir_to_model import parse_directory_of_yaml_files_to_model
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.test.fixtures.id_fixtures import IdNumberSpace, patch_id_generators_helper
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState

logger = logging.getLogger(__name__)


def _data_set_to_read_nodes(
    data_sets: OrderedDict[str, DataSourceDataSet]
) -> OrderedDict[str, ReadSqlSourceNode[DataSourceDataSet]]:
    """Return a mapping from the name of the data source to the dataflow plan node that reads from it."""
    return_dict: OrderedDict[str, ReadSqlSourceNode[DataSourceDataSet]] = OrderedDict()
    for data_source_name, data_set in data_sets.items():
        return_dict[data_source_name] = ReadSqlSourceNode[DataSourceDataSet](data_set)
        logger.debug(f"For data source {data_source_name}, creating node_id {return_dict[data_source_name].node_id}")

    return return_dict


def _data_set_to_source_nodes(
    semantic_model: SemanticModel, data_sets: OrderedDict[str, DataSourceDataSet]
) -> Sequence[BaseOutput[DataSourceDataSet]]:
    source_node_builder = SourceNodeBuilder(semantic_model)
    return source_node_builder.create_from_data_sets(list(data_sets.values()))


@dataclass(frozen=True)
class ConsistentIdObjectRepository:
    """Stores all objects that should have consistent IDs in tests."""

    simple_model_data_sets: OrderedDict[str, DataSourceDataSet]
    simple_model_read_nodes: OrderedDict[str, ReadSqlSourceNode[DataSourceDataSet]]
    simple_model_source_nodes: Sequence[BaseOutput[DataSourceDataSet]]

    multihop_model_read_nodes: OrderedDict[str, ReadSqlSourceNode[DataSourceDataSet]]
    multihop_model_source_nodes: Sequence[BaseOutput[DataSourceDataSet]]

    composite_model_read_nodes: OrderedDict[str, ReadSqlSourceNode[DataSourceDataSet]]
    composite_model_source_nodes: Sequence[BaseOutput[DataSourceDataSet]]


@pytest.fixture(scope="session")
def consistent_id_object_repository(
    simple_semantic_model: SemanticModel,
    multi_hop_join_semantic_model: SemanticModel,
    composite_identifier_semantic_model: SemanticModel,
) -> ConsistentIdObjectRepository:  # noqa: D
    """Create objects that have incremental numeric IDs with a consistent value.

    This should use IDs with a high enough value so that when other tests run with ID generators set to 0 at the start
    of the test and create objects, there is no overlap in the IDs.
    """

    with patch_id_generators_helper(start_value=IdNumberSpace.CONSISTENT_ID_REPOSITORY):
        sm_data_sets = create_data_sets(simple_semantic_model)
        multihop_data_sets = create_data_sets(multi_hop_join_semantic_model)
        composite_data_sets = create_data_sets(composite_identifier_semantic_model)

        return ConsistentIdObjectRepository(
            simple_model_data_sets=sm_data_sets,
            simple_model_read_nodes=_data_set_to_read_nodes(sm_data_sets),
            simple_model_source_nodes=_data_set_to_source_nodes(simple_semantic_model, sm_data_sets),
            multihop_model_read_nodes=_data_set_to_read_nodes(multihop_data_sets),
            multihop_model_source_nodes=_data_set_to_source_nodes(multi_hop_join_semantic_model, multihop_data_sets),
            composite_model_read_nodes=_data_set_to_read_nodes(composite_data_sets),
            composite_model_source_nodes=_data_set_to_source_nodes(
                composite_identifier_semantic_model, composite_data_sets
            ),
        )


def create_data_sets(multihop_semantic_model: SemanticModel) -> OrderedDict[str, DataSourceDataSet]:
    """Convert the DataSources in the model to SqlDataSets.

    Key is the name of the data source, value is the associated data set.
    """
    # Use ordered dict and sort by name to get consistency when running tests.
    data_sets = OrderedDict()
    data_sources: List[DataSource] = multihop_semantic_model.user_configured_model.data_sources
    data_sources.sort(key=lambda x: x.name)

    converter = DataSourceToDataSetConverter(
        column_association_resolver=DefaultColumnAssociationResolver(multihop_semantic_model)
    )

    for data_source in data_sources:
        data_sets[data_source.name] = converter.create_sql_source_data_set(data_source)

    return data_sets


@pytest.fixture(scope="session")
def template_mapping(mf_test_session_state: MetricFlowTestSessionState) -> Dict[str, str]:
    """Mapping for template variables in the model YAML files."""
    schema = mf_test_session_state.mf_source_schema
    return {
        "bookings_source_query": f"SELECT * FROM {schema}.fct_bookings_dt ",
        "bookings_source_table": f"{schema}.fct_bookings",
        "views_source_table": f"{schema}.fct_views",
        "listings_latest_table": f"{schema}.dim_listings_latest",
        "listings_latest": f"{schema}.dim_listings_latest_table",
        "dim_listings_latest_table": f"{schema}.dim_listings_latest",
        "users_latest_table": f"{schema}.dim_users_latest",
        "dim_users_table": f"{schema}.dim_users",
        "fct_id_verifications_table": f"{schema}.fct_id_verifications",
        "fct_revenue_table": f"{schema}.fct_revenue",
        "dim_lux_listing_id_mapping_table": f"{schema}.dim_lux_listing_id_mapping",
        "thorium_table": f"{schema}.thorium",
        "osmium_table": f"{schema}.osmium",
        "dysprosium_table": f"{schema}.dysprosium",
        "dim_companies_table": f"{schema}.dim_companies",
        "source_schema": schema,
        "accounts_source_table": f"{schema}.fct_accounts",
    }


@pytest.fixture(scope="session")
def simple_semantic_model_non_ds(template_mapping: Dict[str, str]) -> SemanticModel:  # noqa: D
    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/non_ds_model"), template_mapping=template_mapping
    )
    return SemanticModel(model_build_result.model)


@pytest.fixture(scope="session")
def simple_semantic_model(template_mapping: Dict[str, str]) -> SemanticModel:  # noqa: D
    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/simple_model"), template_mapping=template_mapping
    )
    return SemanticModel(model_build_result.model)


@pytest.fixture(scope="session")
def composite_identifier_semantic_model(  # noqa: D
    template_mapping: Dict[str, str], mf_test_session_state: MetricFlowTestSessionState
) -> SemanticModel:
    schema = mf_test_session_state.mf_source_schema
    template_mapping = {
        "fct_messages_table": f"{schema}.fct_messages",
        "fct_users_table": f"{schema}.fct_users",
        "fct_users_more_table": f"{schema}.fct_users_more",
    }
    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/composite_identifier_model"),
        template_mapping=template_mapping,
    )
    return SemanticModel(model_build_result.model)


@pytest.fixture(scope="session")
def multi_hop_join_semantic_model(mf_test_session_state: MetricFlowTestSessionState) -> SemanticModel:  # noqa: D
    schema = mf_test_session_state.mf_source_schema
    template_mapping = {
        "account_month_txns": f"{schema}.account_month_txns",
        "customer_table": f"{schema}.customer_table",
        "third_hop_table": f"{schema}.third_hop_table",
        "customer_other_data": f"{schema}.customer_other_data",
        "bridge_table": f"{schema}.bridge_table",
    }
    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/multi_hop_join_model/partitioned_data_sources"),
        template_mapping=template_mapping,
    )
    return SemanticModel(model_build_result.model)


@pytest.fixture(scope="session")
def unpartitioned_multi_hop_join_semantic_model(  # noqa: D
    mf_test_session_state: MetricFlowTestSessionState,
) -> SemanticModel:
    schema = mf_test_session_state.mf_source_schema
    template_mapping = {
        "account_month_txns": f"{schema}.account_month_txns",
        "customer_table": f"{schema}.customer_table",
        "third_hop_table": f"{schema}.third_hop_table",
        "customer_other_data": f"{schema}.customer_other_data",
        "bridge_table": f"{schema}.bridge_table",
    }
    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/multi_hop_join_model/unpartitioned_data_sources"),
        template_mapping=template_mapping,
    )
    return SemanticModel(model_build_result.model)


@pytest.fixture(scope="session")
def simple_user_configured_model(template_mapping: Dict[str, str]) -> UserConfiguredModel:
    """Model used for many tests."""

    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/simple_model"), template_mapping=template_mapping
    )
    return model_build_result.model


@pytest.fixture(scope="session")
def simple_model__pre_transforms(template_mapping: Dict[str, str]) -> UserConfiguredModel:
    """Model used for tests pre-transformations."""

    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/simple_model"),
        template_mapping=template_mapping,
        apply_pre_transformations=True,
        apply_post_transformations=False,
    )
    return model_build_result.model


@pytest.fixture(scope="session")
def extended_date_semantic_model(mf_test_session_state: MetricFlowTestSessionState) -> SemanticModel:  # noqa: D
    schema = mf_test_session_state.mf_source_schema
    template_mapping = {
        "fct_bookings_extended_table": f"{schema}.fct_bookings_extended",
        "fct_bookings_extended_monthly_table": f"{schema}.fct_bookings_extended_monthly",
        "dim_listings_extended_table": f"{schema}.dim_listings_extended",
    }
    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/extended_date_model"),
        template_mapping=template_mapping,
    )
    return SemanticModel(model_build_result.model)


@pytest.fixture(scope="session")
def data_warehouse_validation_model(template_mapping: Dict[str, str]) -> UserConfiguredModel:
    """Model used for data warehouse validation tests"""

    model_build_result = parse_directory_of_yaml_files_to_model(
        os.path.join(os.path.dirname(__file__), "model_yamls/data_warehouse_validation_model"),
        template_mapping=template_mapping,
    )
    return model_build_result.model


@pytest.fixture(scope="session")
def config_linter_model_path() -> str:
    """Model path for ConfigLinter tests"""

    return os.path.join(os.path.dirname(__file__), "model_yamls/config_linter_model")



import copy
import pytest
from typing import Tuple

from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import DataSourceElementType, ModelValidationException
from metricflow.test.test_utils import find_data_source_with


def _categorical_dimensions(data_source: DataSource) -> Tuple[Dimension, ...]:
    return tuple(dim for dim in data_source.dimensions if dim.type == DimensionType.CATEGORICAL)


def test_cross_element_names(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa:D
    model = copy.deepcopy(simple_model__pre_transforms)

    # ensure we have a usable data source for the test
    usable_ds, usable_ds_index = find_data_source_with(
        model,
        lambda data_source: len(data_source.measures) > 0
        and len(data_source.identifiers) > 0
        and len(_categorical_dimensions(data_source=data_source)) > 0,
    )

    measure_reference = usable_ds.measures[0].reference
    # If the matching dimension is a time dimension we can accidentally create two primary time dimensions, and then
    # the validation will throw a different error and fail the test
    dimension_reference = _categorical_dimensions(data_source=usable_ds)[0].reference

    ds_measure_x_dimension = copy.deepcopy(usable_ds)
    ds_measure_x_identifier = copy.deepcopy(usable_ds)
    ds_dimension_x_identifier = copy.deepcopy(usable_ds)

    # We update the matching categorical dimension by reference for convenience
    ds_measure_x_dimension.get_dimension(dimension_reference).name = measure_reference.element_name
    ds_measure_x_identifier.identifiers[0].name = measure_reference.element_name
    ds_dimension_x_identifier.identifiers[0].name = dimension_reference.element_name

    model.data_sources[usable_ds_index] = ds_measure_x_dimension
    with pytest.raises(
        ModelValidationException,
        match=(
            f"element `{measure_reference.element_name}` is of type {DataSourceElementType.DIMENSION}, but it is used as "
            f"types .*?DataSourceElementType.DIMENSION.*?DataSourceElementType.MEASURE.*? across the model"
        ),
    ):
        ModelValidator().checked_validations(model)

    model.data_sources[usable_ds_index] = ds_measure_x_identifier
    with pytest.raises(
        ModelValidationException,
        match=(
            f"element `{measure_reference.element_name}` is of type {DataSourceElementType.IDENTIFIER}, but it is used as "
            f"types .*?DataSourceElementType.IDENTIFIER.*?DataSourceElementType.MEASURE.*? across the model"
        ),
    ):
        ModelValidator().checked_validations(model)

    model.data_sources[usable_ds_index] = ds_dimension_x_identifier
    with pytest.raises(
        ModelValidationException,
        match=(
            f"element `{dimension_reference.element_name}` is of type {DataSourceElementType.DIMENSION}, but it is used as "
            f"types .*?DataSourceElementType.DIMENSION.*?DataSourceElementType.IDENTIFIER.*? across the model"
        ),
    ):
        ModelValidator().checked_validations(model)



from collections import defaultdict
from dataclasses import dataclass
from operator import xor
import traceback
from typing import DefaultDict, Dict, List, Optional, Set
from dbt.contracts.graph.parsed import ParsedMetric as DbtMetric, ParsedModelNode as DbtModelNode
from dbt.contracts.graph.unparsed import MetricFilter as DbtMetricFilter
from dbt.exceptions import ref_invalid_args
from dbt.contracts.graph.manifest import Manifest as DbtManifest
from metricflow.aggregation_properties import AggregationType
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType, DimensionTypeParams
from metricflow.model.objects.elements.identifier import Identifier
from metricflow.model.objects.elements.measure import Measure
from metricflow.model.objects.metric import Metric, MetricInputMeasure, MetricType, MetricTypeParams
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.parsing.dir_to_model import ModelBuildResult
from metricflow.model.validations.validator_helpers import ModelValidationResults, ValidationError, ValidationIssue
from metricflow.time.time_granularity import TimeGranularity


@dataclass
class TransformedDbtMetric:  # noqa: D
    data_source: DataSource
    metric: Metric


CALC_METHOD_TO_MEASURE_TYPE: Dict[str, AggregationType] = {
    "count": AggregationType.COUNT,
    "count_distinct": AggregationType.COUNT_DISTINCT,
    "sum": AggregationType.SUM,
    "average": AggregationType.AVERAGE,
    "min": AggregationType.MIN,
    "max": AggregationType.MAX,
    # "derived": AggregationType.DERIVED # Derived DBT metrics don't create measures
}


class DbtManifestTransformer:
    """The DbtManifestTransform is a class used to transform dbt Manifests into MetricFlow UserConfiguredModels

    This helps keep track of state objects while transforming the Manifest into a
    UserConfiguredModel, ensuring like dbt Node elements are rendered only once and
    allowing us to pass around fewer arguments (reducing the mental load)
    """

    def __init__(self, manifest: DbtManifest) -> None:
        """Constructor.

        Args:
            manifest: A dbt Manifest object
        """
        self.manifest = manifest
        self._time_dimension_stats: Optional[Dict[str, List[str]]] = None  # lazy load it
        self._resolved_dbt_model_refs: Dict[int, DbtModelNode] = {}  # cache of resolved nodes

    @property
    def time_dimension_stats(self) -> Dict[str, List[str]]:
        """The stats on time dimensions from the dbt Manifest

        The time dimension stats are returned as a dictionary where in the keys
        are the names of the time dimension, and the value associated with each
        key is a list of `DbtMetric.model`s which that the associated time
        dimension should be primary for
        """
        if not self._time_dimension_stats:
            self._time_dimension_stats = self.collect_time_dimension_stats(dbt_metrics=self.manifest.metrics.values())

        return self._time_dimension_stats

    def resolve_metric_model(self, dbt_metric: DbtMetric) -> DbtModelNode:
        """Returns a DbtModelNode based on the `DbtMetric.model`

        `DbtMetric.model` string values should either be a model name, or a
        dbt [ref] (https://docs.getdbt.com/reference/dbt-jinja-functions/ref) object for a model. A dbt `ref` contains either one or two
        strings. If two strings are specified, the first is the target package
        and the second is the target model. If only one string is specified
        then it represents the target model.

        TODO: There should be a way to do this using dbt, and ideally we'll
        transition to doing this whatever that way is, but for the time being
        this was the fastest solution time wise. In addition to resolving the
        ref, this method caches the resolved ref node and uses the cache when
        possible to avoid re-resolving nodes.
        """
        if dbt_metric.model is None:
            raise RuntimeError("Unable to resolve a model for a `DbtMetric.model` value of `None`")

        target_model = None
        target_package = None

        if dbt_metric.model[:4] == "ref(":
            # Parse the DbtMetric.model into it's parts
            ref_parts = dbt_metric.model[4:-1].split(",")
            if len(ref_parts) == 1:
                target_model = ref_parts[0].strip(" \"'\t\r\n")
            elif len(ref_parts) == 2:
                target_package = ref_parts[0].strip(" \"'\t\r\n")
                target_model = ref_parts[1].strip(" \"'\t\r\n")
            else:
                ref_invalid_args(dbt_metric.name, ref_parts)
        else:
            target_model = dbt_metric.model

        hashed = hash((target_model, target_package, self.manifest.metadata.project_id, dbt_metric.package_name))
        if hashed not in self._resolved_dbt_model_refs:
            node = self.manifest.resolve_ref(
                target_model_name=target_model,
                target_model_package=target_package,
                current_project=self.manifest.metadata.project_id,
                node_package=dbt_metric.package_name,
            )
            assert isinstance(
                node, DbtModelNode
            ), f"Ref `{dbt_metric.model}` resolved to {node}, which is not of type `{DbtModelNode.__name__}`"

            self._resolved_dbt_model_refs[hashed] = node

        return self._resolved_dbt_model_refs[hashed]

    @classmethod
    def db_table_from_model_node(cls, node: DbtModelNode) -> str:
        """Get the '.' joined database table name of a DbtModelNode"""
        return f"{node.database}.{node.schema}.{node.name}"

    def _build_dimension(self, name: str, dbt_metric: DbtMetric) -> Dimension:
        """Helper for `build_dimenions which builds either a categorical or time dimension"""

        if name in self.time_dimension_stats.keys():
            return Dimension(
                name=name,
                type=DimensionType.TIME,
                type_params=DimensionTypeParams(
                    is_primary=dbt_metric.model in self.time_dimension_stats[name], time_granularity=TimeGranularity.DAY
                ),
            )
        else:
            return Dimension(
                name=name,
                type=DimensionType.CATEGORICAL,
            )

    def build_dimensions(self, dbt_metric: DbtMetric) -> List[Dimension]:
        """Given a DbtMetric, builds all the associated MetricFlow dimensions"""
        dimensions = []

        # Build dimensions specifically from DbtMetric.dimensions list
        for dimension in dbt_metric.dimensions:
            dimensions.append(self._build_dimension(name=dimension, dbt_metric=dbt_metric))

        # Add DbtMetric.timestamp as a time dimension
        dimensions.append(self._build_dimension(name=dbt_metric.timestamp, dbt_metric=dbt_metric))

        # We need to deduplicate the filters because a field could be the same in
        # two filters. For example, if two filters exist for `amount`, one with
        # `>= 500` and the other `< 1000`, but only one dimension should be created
        distinct_dbt_metric_filter_fields = set([filter.field for filter in dbt_metric.filters])
        # Add dimension per distinct filter field
        # exclude when field is also listed as a DbtMetric.dimension
        # exclude when field is also the DbtMetric.timestamp
        for filter_field in distinct_dbt_metric_filter_fields:
            if filter_field not in dbt_metric.dimensions and filter_field != dbt_metric.timestamp:
                dimensions.append(self._build_dimension(name=filter_field, dbt_metric=dbt_metric))

        return dimensions

    def build_measure(self, dbt_metric: DbtMetric) -> Measure:
        """Attemps to build a measure for a given DbtMetric

        Raises:
            RuntimeError: A measure can't be built for `derived` dbt metrics
        """
        if dbt_metric.calculation_method == "derived":
            raise RuntimeError("Cannot build a MetricFlow measure for `derived` DbtMetric")

        return Measure(
            name=dbt_metric.name,
            agg=CALC_METHOD_TO_MEASURE_TYPE[dbt_metric.calculation_method],
            expr=dbt_metric.expression,
            agg_time_dimension=dbt_metric.timestamp,
        )

    def build_data_source_for_metric(self, dbt_metric: DbtMetric) -> DataSource:
        """Attemps to build a data source for a given DbtMetric

        Raises:
            RuntimeError: A data source can't be built for `derived` dbt metrics
        """
        if dbt_metric.calculation_method == "derived":
            raise RuntimeError("Cannot build a MetricFlow data source for `derived` DbtMetric")

        metric_model_ref = self.resolve_metric_model(dbt_metric=dbt_metric)
        data_source_table = self.db_table_from_model_node(metric_model_ref)
        return DataSource(
            name=metric_model_ref.name,
            description=metric_model_ref.description,
            sql_table=data_source_table,
            dbt_model=data_source_table,
            dimensions=self.build_dimensions(dbt_metric),
            measures=[self.build_measure(dbt_metric)],
        )

    @classmethod
    def build_where_stmt_from_filters(cls, filters: List[DbtMetricFilter]) -> str:
        """Builds an SQL 'where' statement from the passed in filters

        Each dbt filter has a field, an operator, and a value. With these dbt
        forms the individual statment '{field} {operator} {value}' and joins
        them with an 'AND'. Thus we do the same.

        Note:
            TODO: We could probably replace this with whatever method dbt uses to
            build the statement.
        """
        clauses = [f"{filter.field} {filter.operator} {filter.value}" for filter in filters]
        return " AND ".join(clauses)

    def build_proxy_metric(self, dbt_metric: DbtMetric) -> Metric:
        """Attempt to build a proxy metric for the given DbtMetric

        For DbtMetrics which have a calculation method != 'derived',
        we have separately created a measure of the appropriate type.
        This method creates the proxy metric for the measure.

        Raises:
            RuntimeError: A proxy metric can't be built for `derived` dbt metrics
        """
        if dbt_metric.calculation_method == "derived":
            raise RuntimeError("Cannot build a MetricFlow proxy metric for `derived` DbtMetric")

        where_clause_constraint: Optional[WhereClauseConstraint] = None
        if dbt_metric.filters:
            where_clause_constraint = WhereClauseConstraint(
                where=self.build_where_stmt_from_filters(filters=dbt_metric.filters),
                linkable_names=[filter.field for filter in dbt_metric.filters],
            )

        return Metric(
            name=dbt_metric.name,
            description=dbt_metric.description,
            type=MetricType.MEASURE_PROXY,
            type_params=MetricTypeParams(
                measure=MetricInputMeasure(name=dbt_metric.name),
            ),
            constraint=where_clause_constraint,
        )

    def dbt_metric_to_metricflow_elements(self, dbt_metric: DbtMetric) -> TransformedDbtMetric:
        """Builds a MetricFlow data source and proxy metric for the given DbtMetric"""
        data_source = self.build_data_source_for_metric(dbt_metric)
        proxy_metric = self.build_proxy_metric(dbt_metric)
        return TransformedDbtMetric(data_source=data_source, metric=proxy_metric)

    @classmethod
    def deduplicate_data_sources(cls, data_sources: List[DataSource]) -> DataSource:
        """Attempts to deduplicate a list of data sources into a single data source

        Because each DbtMetric (which isn't `derived`) creates a data source,
        and many DbtMetric can create the same data source with differring
        defintions for dimensions and measures, we need a way to deduplicate/merge
        them. This function does that. It requires that the base information
        (name/table/query/description/etc) of the data source not be variable and
        that dimensions/measures/identifers with the same name have the same
        attributes.
        """

        if len(data_sources) == 1:
            return data_sources[0]

        # collect the variations of data source properties
        measures: Set[Measure] = set()
        identifiers: Set[Identifier] = set()
        dimensions: Set[Dimension] = set()
        names: Set[str] = set()
        descriptions: Set[str] = set()
        sql_tables: Set[str] = set()
        sql_queries: Set[str] = set()
        dbt_models: Set[str] = set()
        for data_source in data_sources:
            # This is an atypical pattern but results in less work. The following
            # five lines are ternaries wherein the attribute is added to tracking
            # set for the attribute, but only if the attribute is defined for the
            # data source. The `else None` is required because python ternaries
            # require an `else` statement. Having the else be `None` is simply
            # saying nothing happens, i.e. the set is not modified.
            names.add(data_source.name) if data_source.name else None
            descriptions.add(data_source.description) if data_source.description else None
            sql_tables.add(data_source.sql_table) if data_source.sql_table else None
            sql_queries.add(data_source.sql_query) if data_source.sql_query else None
            dbt_models.add(data_source.dbt_model) if data_source.dbt_model else None

            # ensure any unique sub elements get added to the set of sub elements
            measures = measures.union(set(data_source.measures)) if data_source.measures else measures
            identifiers = identifiers.union(set(data_source.identifiers)) if data_source.identifiers else identifiers
            dimensions = dimensions.union(set(data_source.dimensions)) if data_source.dimensions else dimensions

        assert len(names) == 1, "Cannot merge data sources, all data sources to merge must have same name"
        assert (
            len(descriptions) <= 1
        ), "Cannot merge data sources, all data sources to merge must have same descritpion (or none)"
        assert (
            len(sql_tables) <= 1
        ), "Cannot merge data sources, all data sources to merge must have same sql_table (or none)"
        assert (
            len(sql_queries) <= 1
        ), "Cannot merge data sources, all data sources to merge must have same sql_query (or none)"
        assert xor(
            len(sql_tables) == 1, len(sql_queries) == 1
        ), "Cannot merge data sources, definitions for both sql_table and sql_query exist"
        assert (
            len(dbt_models) <= 1
        ), "Cannot merge data sources, all data sources to merge must have same dbt_model (or none)"

        return DataSource(
            name=list(names)[0],
            description=list(descriptions)[0] if descriptions else None,
            sql_table=list(sql_tables)[0] if sql_tables else None,
            sql_query=list(sql_queries)[0] if sql_queries else None,
            dbt_model=list(dbt_models)[0] if dbt_models else None,
            dimensions=list(dimensions),
            identifiers=list(identifiers),
            measures=list(measures),
        )

    @classmethod
    def collect_time_dimension_stats(cls, dbt_metrics: List[DbtMetric]) -> Dict[str, List[str]]:
        """Compiles stats on time dimensions for a given list of DbtMetrics

        Arguably this is probably a fairly black magic function. This function
        determines which time dimensions are primary for which `DbtMetric.model`
        """
        # For storing the time dimension names and the `DbtMetric.model`s they are primary for
        time_dimensions: Dict[str, List[str]] = {}

        # Map each `DbtMetric.model` value to counts of how many times each time
        # dimension is associated with it. The time dimension that is associated
        # with a `DbtMetric.model` the most will be considered the primary time
        # dimenson for the data source that is built for the `DbtMetric.model`
        time_stats_for_metric_models: Dict[str, Dict[str, int]] = {}
        for dbt_metric in dbt_metrics:
            if dbt_metric.calculation_method != "derived":
                if dbt_metric.timestamp not in time_dimensions:
                    time_dimensions[dbt_metric.timestamp] = []

                if dbt_metric.model not in time_stats_for_metric_models:
                    time_stats_for_metric_models[dbt_metric.model] = {dbt_metric.timestamp: 1}
                else:
                    time_stats_for_metric_models[dbt_metric.model][dbt_metric.timestamp] += 1

        # Take the mapping created above and set the `DbtMetric.model` values
        # that should be primary for the time dimension
        for metric_model, time_stats in time_stats_for_metric_models.items():
            primary_time_dim = max(time_stats, key=time_stats.get)  # type: ignore
            time_dimensions[primary_time_dim].append(metric_model)

        return time_dimensions

    def build_user_configured_model(self) -> ModelBuildResult:
        """Builds a UserConfiguredModel from the manifest of the instance

        Note:
            TODO: This currently skips DbtMetric that are `derived`. Once MetricFlow
            supports derived metrics, we'll need to add that functionality to
            handle dbt derived metrics -> metricflow derived metrics.
        """
        data_sources_map: DefaultDict[str, List[DataSource]] = defaultdict(list)
        metrics = []
        issues: List[ValidationIssue] = []

        for dbt_metric in self.manifest.metrics.values():
            # TODO: Handle derived dbt metrics
            if dbt_metric.calculation_method == "derived":
                continue
            else:
                transformed_dbt_metric = self.dbt_metric_to_metricflow_elements(dbt_metric=dbt_metric)
                data_sources_map[transformed_dbt_metric.data_source.name].append(transformed_dbt_metric.data_source)
                metrics.append(transformed_dbt_metric.metric)

        # As it might be the case that we generated many of the same data source,
        # we need to merge / dedupe them
        deduped_data_sources = []
        for name, data_sources in data_sources_map.items():
            try:
                deduped_data_sources.append(self.deduplicate_data_sources(data_sources))
            except Exception as e:
                issues.append(
                    ValidationError(
                        message=f"Failed to merge data sources with the name `{name}`",
                        extra_detail="".join(traceback.format_tb(e.__traceback__)),
                    )
                )

        return ModelBuildResult(
            model=UserConfiguredModel(data_sources=list(deduped_data_sources), metrics=metrics),
            issues=ModelValidationResults.from_issues_sequence(issues=issues),
        )



import os
import pathlib

from metricflow.configuration.constants import CONFIG_PATH_KEY
from metricflow.configuration.yaml_handler import YamlFileHandler


class ConfigHandler(YamlFileHandler):
    """Class to handle all config file retrieval/insertion actions."""

    def __init__(self) -> None:  # noqa: D
        # Create config directory if not exist
        if not os.path.exists(self.dir_path):
            dir_path = pathlib.Path(self.dir_path)
            dir_path.mkdir(parents=True)
        super().__init__(yaml_file_path=self.file_path)

    @property
    def dir_path(self) -> str:
        """Retrieve MetricFlow config directory from $MF_CONFIG_DIR, default config dir is ~/.metricflow"""
        config_dir_env = os.getenv(CONFIG_PATH_KEY)
        return config_dir_env if config_dir_env else f"{str(pathlib.Path.home())}/.metricflow"

    @property
    def file_path(self) -> str:
        """Config file can be found at <config_dir>/config.yml"""
        return os.path.join(self.dir_path, "config.yml")

    @property
    def log_file_path(self) -> str:
        """Returns the name of the log file where all logging messages are stored."""
        log_dir_elements = [self.dir_path, "logs"]
        log_dir = os.path.join(*log_dir_elements)
        os.makedirs(log_dir, exist_ok=True)
        return os.path.join(*(log_dir_elements + ["metricflow.log"]))



import os
from typing import List, Optional
import yaml
from yamllint import config, linter, rules
from metricflow.model.parsing.dir_to_model import (
    METRIC_TYPE,
    DATA_SOURCE_TYPE,
    MATERIALIZATION_TYPE,
    collect_yaml_config_file_paths,
)

from metricflow.model.validations.validator_helpers import (
    FileContext,
    ModelValidationResults,
    ValidationError,
    ValidationIssue,
    ValidationWarning,
)

WARNING = "warning"
ERROR = "error"
DISABLE = "disable"
LEVEL = "level"


class ConfigLinter:  # noqa: D
    """A linter for checking the format of MetricFlow YAML model config files

    This config linter is a custom implementation of the linter provided by
    the yamllint package (https://pypi.org/project/yamllint/). For any problems
    that are found, ValidationIssues are created (with the appropriate level).
    Breaking issues will return errors. Issues that won't cause problems outside
    of maintainability will be warnings. A custom config can be provided on
    initialization. Please see the yamllint documentation for how that can be
    provided (https://yamllint.readthedocs.io/en/stable/configuration.html#custom-configuration-without-a-config-file)
    """

    DEFAULT_CONFIG = {
        "yaml-files": ["*.yaml", "*.yml"],
        "rules": {
            rules.braces.ID: DISABLE,
            rules.brackets.ID: DISABLE,
            rules.colons.ID: DISABLE,
            rules.commas.ID: DISABLE,
            rules.comments.ID: DISABLE,
            rules.comments_indentation.ID: DISABLE,  # TODO: Turn on as warning once warnings are less noisy in cli
            rules.document_start.ID: DISABLE,  # TODO: Turn on as warning once warnings are less noisy in cli
            rules.document_end.ID: DISABLE,
            rules.empty_lines.ID: DISABLE,
            rules.empty_values.ID: {LEVEL: ERROR},
            rules.hyphens.ID: DISABLE,
            rules.indentation.ID: DISABLE,  # TODO: Turn on as warning once warnings are less noisy in cli
            rules.key_duplicates.ID: {LEVEL: ERROR},
            rules.key_ordering.ID: DISABLE,
            rules.line_length.ID: DISABLE,
            rules.new_line_at_end_of_file.ID: DISABLE,
            rules.new_lines.ID: DISABLE,
            rules.octal_values.ID: DISABLE,  # TODO: Turn on as warning once warnings are less noisy in cli
            rules.quoted_strings.ID: DISABLE,
            rules.trailing_spaces.ID: DISABLE,  # TODO: Turn on as warning once warnings are less noisy in cli
            rules.truthy.ID: DISABLE,
        },
    }

    def __init__(self, lint_config: Optional[str] = None) -> None:  # noqa: D
        """Constructor for building a ConfigLinter to then use with MetricFlow YAML model config files

        Args:
            lint_config: A JSON-ified dict representing a yamllint config, see
                https://yamllint.readthedocs.io/en/stable/configuration.html#custom-configuration-without-a-config-file
                for details. The default config can be found as `ConfigLinter.DEFAULT_CONFIG`
        """
        self._lint_config = lint_config if lint_config else yaml.dump(data=self.DEFAULT_CONFIG)
        self._config = config.YamlLintConfig(self._lint_config)

    @staticmethod
    def add_additional_key_duplicates_info(problem: linter.LintProblem) -> None:
        """Add additional information to the key_duplicates problem description"""

        # desc is of the form f'duplication of key "{key}" in mapping'
        # the following split gives us ['duplication of key ', <key>, ' in mapping']
        key = problem.desc.split('"')[1]
        if key in [DATA_SOURCE_TYPE, MATERIALIZATION_TYPE, METRIC_TYPE]:
            problem.desc += f'. Key "{key}" is a top level object. If multiple top level objects are specified in the same file, they must be separated using "---"'

    def lint_file(self, file_path: str) -> List[ValidationIssue]:  # noqa: D
        """Lints a given YAML file for structural issues, returning a list of ValidationIssues"""
        (_head, file_name) = os.path.split(file_path)

        issues: List[ValidationIssue] = []
        with open(file_path) as f:
            for problem in linter.run(f, self._config):

                if problem.rule == rules.key_duplicates.ID:
                    self.add_additional_key_duplicates_info(problem=problem)

                if problem.level == ERROR:
                    issues.append(
                        ValidationError(
                            context=FileContext(file_name=file_name, line_number=problem.line),
                            message=problem.desc,  # type: ignore[misc]
                        )
                    )
                else:
                    issues.append(
                        ValidationWarning(
                            context=FileContext(file_name=file_name, line_number=problem.line),
                            message=problem.desc,  # type: ignore[misc]
                        )
                    )

        return issues

    def lint_files(self, file_paths: List[str]) -> ModelValidationResults:
        """Gets a list of ValidationIssues for the given list of YAML file paths"""
        issues: List[ValidationIssue] = []
        for file_path in file_paths:
            issues += self.lint_file(file_path)

        return ModelValidationResults.from_issues_sequence(issues=issues)

    def lint_dir(self, dir_path: str) -> ModelValidationResults:
        """Generates ModelValidationResults object for lint issues of model config files found in the given directory"""
        config_file_paths = collect_yaml_config_file_paths(directory=dir_path)
        return self.lint_files(file_paths=config_file_paths)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List

from metricflow.inference.models import InferenceSignal
from metricflow.inference.context.data_warehouse import DataWarehouseInferenceContext


class InferenceRule(ABC):
    """Implements some sort of heuristic that produces signals about columns.

    An inference rule produces zero or more `InferenceSignal` instances about whatever
    columns it thinks it should, based on input `InferenceContext`s.

    Concrete implementations should aim to be short and modularized. It is preferred to
    compose multiple small rules that each produce a signal type than to make one large
    rule with complex logic to produce a bunch of signals.

    It currently only accepts DataWarehouseInferenceContext as input, but this will probably
    be generalized later.
    """

    @abstractmethod
    def process(self, warehouse: DataWarehouseInferenceContext) -> List[InferenceSignal]:
        """The actual rule implementation that returns a list of signals based on the input contexts."""
        raise NotImplementedError



import copy
import re

import pytest
from metricflow.aggregation_properties import AggregationType
from metricflow.model.objects.elements.dimension import Dimension, DimensionType, DimensionTypeParams
from metricflow.model.objects.elements.measure import Measure, NonAdditiveDimensionParameters
from metricflow.model.objects.metric import Metric, MetricInputMeasure, MetricType, MetricTypeParams
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.model_validator import ModelValidator
from metricflow.model.validations.validator_helpers import ModelValidationException
from metricflow.object_utils import flatten_nested_sequence
from metricflow.test.test_utils import find_data_source_with
from metricflow.time.time_granularity import TimeGranularity


def test_metric_missing_measure(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests the basic MetricMeasuresRule, which asserts all measure inputs to a metric exist in the model"""
    model = copy.deepcopy(simple_model__pre_transforms)
    metric_name = "invalid_measure_metric_do_not_add_to_model"
    measure_name = "this_measure_cannot_exist_or_else_it_breaks_tests"

    with pytest.raises(
        ModelValidationException,
        match=f"Measure {measure_name} referenced in metric {metric_name} is not defined in the model!",
    ):
        model.metrics.append(
            Metric(
                name=metric_name,
                description="Metric with invalid measure",
                type=MetricType.EXPR,
                type_params=MetricTypeParams(measures=[MetricInputMeasure(name=measure_name)]),
            )
        )

        ModelValidator().checked_validations(model=model)


def test_measures_only_exist_in_one_data_source(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copy.deepcopy(simple_model__pre_transforms)

    build = ModelValidator().validate_model(model)
    duplicate_measure_message = "Found measure with name .* in multiple data sources with names"
    found_issue = False

    if build.issues is not None:
        for issue in build.issues.all_issues:
            if re.search(duplicate_measure_message, issue.message):
                found_issue = True

    assert found_issue is False

    # add measure present in one data source to another
    first_data_source, _ = find_data_source_with(
        model, lambda data_source: data_source.measures is not None and len(data_source.measures) > 0
    )

    second_data_source, _ = find_data_source_with(
        model, lambda data_source: data_source.measures is not None and data_source.name != first_data_source.name
    )

    measure = first_data_source.measures[0]
    second_data_source.measures = list(flatten_nested_sequence([second_data_source.measures, [measure]]))

    build = ModelValidator().validate_model(model)

    if build.issues is not None:
        for issue in build.issues.all_issues:
            if re.search(duplicate_measure_message, issue.message):
                found_issue = True

    assert found_issue is True


def test_measure_alias_is_set_when_required(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests to ensure that an appropriate error appears when a required alias is missing"""
    model = copy.deepcopy(simple_model__pre_transforms)
    measure_name = find_data_source_with(model, lambda data_source: len(data_source.measures) > 0)[0].measures[0].name
    model.metrics.append(
        Metric(
            name="metric1",
            type=MetricType.EXPR,
            type_params=MetricTypeParams(
                measures=[
                    MetricInputMeasure(name=measure_name),
                    MetricInputMeasure(name=measure_name, constraint="is_instant"),
                ]
            ),
        )
    )

    build = ModelValidator().validate_model(model)

    assert len(build.issues.errors) == 1
    expected_error_substring = f"depends on multiple different constrained versions of measure {measure_name}"
    actual_error = build.issues.errors[0].as_readable_str()
    assert (
        actual_error.find(expected_error_substring) != -1
    ), f"Expected error {expected_error_substring} not found in error string! Instead got {actual_error}"


def test_invalid_measure_alias_name(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests measures with aliases that don't pass the unique and valid name rule"""
    model = copy.deepcopy(simple_model__pre_transforms)
    measure_name = find_data_source_with(model, lambda data_source: len(data_source.measures) > 0)[0].measures[0].name
    invalid_alias = "_can't_do_this_"
    model.metrics.append(
        Metric(
            name="metric1",
            type=MetricType.EXPR,
            type_params=MetricTypeParams(
                measures=[MetricInputMeasure(name=measure_name, constraint="is_instant", alias=invalid_alias)]
            ),
        )
    )

    build = ModelValidator().validate_model(model)

    assert len(build.issues.errors) == 1
    expected_error_substring = f"Invalid name `{invalid_alias}` - names should only consist of"
    actual_error = build.issues.errors[0].as_readable_str()
    assert (
        actual_error.find(expected_error_substring) != -1
    ), f"Expected error {expected_error_substring} not found in error string! Instead got {actual_error}"


def test_measure_alias_measure_name_conflict(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests measures with aliases that already exist as measure names"""
    model = copy.deepcopy(simple_model__pre_transforms)
    measures = find_data_source_with(model, lambda data_source: len(data_source.measures) > 1)[0].measures
    measure_name = measures[0].name
    invalid_alias = measures[1].name
    model.metrics.append(
        Metric(
            name="metric1",
            type=MetricType.EXPR,
            type_params=MetricTypeParams(
                measures=[MetricInputMeasure(name=measure_name, constraint="is_instant", alias=invalid_alias)]
            ),
        )
    )

    build = ModelValidator().validate_model(model)

    assert len(build.issues.errors) == 1
    expected_error_substring = (
        f"Alias `{invalid_alias}` for measure `{measure_name}` conflicts with measure names "
        f"defined elsewhere in the model!"
    )
    actual_error = build.issues.errors[0].as_readable_str()
    assert (
        actual_error.find(expected_error_substring) != -1
    ), f"Expected error {expected_error_substring} not found in error string! Instead got {actual_error}"


def test_reused_measure_alias(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests measures with aliases that have been used as measure aliases elsewhere in the model"""
    model = copy.deepcopy(simple_model__pre_transforms)
    measures = find_data_source_with(model, lambda data_source: len(data_source.measures) > 1)[0].measures
    first_measure_name = measures[0].name
    second_measure_name = measures[1].name
    invalid_alias = "duplicate_alias"
    model.metrics.append(
        Metric(
            name="metric1",
            type=MetricType.EXPR,
            type_params=MetricTypeParams(
                measures=[MetricInputMeasure(name=first_measure_name, constraint="is_instant", alias=invalid_alias)]
            ),
        ),
    )
    model.metrics.append(
        Metric(
            name="metric2",
            type=MetricType.EXPR,
            type_params=MetricTypeParams(
                measures=[MetricInputMeasure(name=second_measure_name, constraint="is_instant", alias=invalid_alias)]
            ),
        ),
    )

    build = ModelValidator().validate_model(model)

    assert len(build.issues.errors) == 1
    expected_error_substring = (
        f"Measure alias {invalid_alias} conflicts with a measure alias used elsewhere in the model!"
    )
    actual_error = build.issues.errors[0].as_readable_str()
    assert (
        actual_error.find(expected_error_substring) != -1
    ), f"Expected error {expected_error_substring} not found in error string! Instead got {actual_error}"


def test_reused_measure_alias_within_metric(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests measures with aliases that have been used as measure aliases in the same metric spec

    This covers a boundary case in the logic where alias checking must always include aliases from the current
    metric under consideration, instead of simply checking against the previous seen values
    """
    model = copy.deepcopy(simple_model__pre_transforms)
    measures = find_data_source_with(model, lambda data_source: len(data_source.measures) > 1)[0].measures
    first_measure_name = measures[0].name
    second_measure_name = measures[1].name
    invalid_alias = "duplicate_alias"
    model.metrics.append(
        Metric(
            name="metric1",
            type=MetricType.EXPR,
            type_params=MetricTypeParams(
                measures=[
                    MetricInputMeasure(name=first_measure_name, constraint="is_instant", alias=invalid_alias),
                    MetricInputMeasure(name=second_measure_name, constraint="is_instant", alias=invalid_alias),
                ]
            ),
        ),
    )

    build = ModelValidator().validate_model(model)

    assert len(build.issues.errors) == 1
    expected_error_substring = (
        f"Measure alias {invalid_alias} conflicts with a measure alias used elsewhere in the model!"
    )
    actual_error = build.issues.errors[0].as_readable_str()
    assert (
        actual_error.find(expected_error_substring) != -1
    ), f"Expected error {expected_error_substring} not found in error string! Instead got {actual_error}"


def test_invalid_non_additive_dimension_properties(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests validator for invalid cases of non_additive_dimension properties."""
    model = copy.deepcopy(simple_model__pre_transforms)
    invalid_measure = Measure(
        name="bad_measure",
        agg=AggregationType.SUM,
        non_additive_dimension=NonAdditiveDimensionParameters(
            name="doesntexist",  # Time dimension doesn't exist
            window_choice=AggregationType.SUM,  # Should only be MIN/MAX
            window_groupings=["wherethisidentifier"],  # Identifier doesn't exist
        ),
        agg_time_dimension="ds",
    )
    invalid_measure2 = Measure(
        name="bad_measure2",
        agg=AggregationType.SUM,
        non_additive_dimension=NonAdditiveDimensionParameters(
            name="is_instant",  # Is a CATEGORICAL dimension
            window_choice=AggregationType.MIN,
        ),
        agg_time_dimension="ds",
    )
    invalid_measure3 = Measure(
        name="bad_measure3",
        agg=AggregationType.SUM,
        non_additive_dimension=NonAdditiveDimensionParameters(
            name="weekly_time",  # Not same base granularity as agg_time_dimension
            window_choice=AggregationType.MIN,
        ),
        agg_time_dimension="ds",
    )
    weekly_time_dimension = Dimension(
        name="weekly_time",
        type=DimensionType.TIME,
        type_params=DimensionTypeParams(time_granularity=TimeGranularity.WEEK),
    )
    data_source_with_measures = find_data_source_with(model, lambda data_source: data_source.name == "bookings_source")[
        0
    ]
    data_source_with_measures.dimensions.extend([weekly_time_dimension])  # type: ignore
    data_source_with_measures.measures.extend([invalid_measure, invalid_measure2, invalid_measure3])  # type: ignore
    build = ModelValidator().validate_model(model)
    expected_error_substring_1 = (
        f"that is not defined as a dimension in data source '{data_source_with_measures.name}'."
    )
    expected_error_substring_2 = "has a non_additive_dimension with an invalid 'window_choice'"
    expected_error_substring_3 = "has a non_additive_dimension with an invalid 'window_groupings'"
    expected_error_substring_4 = "that is defined as a categorical dimension which is not supported."
    expected_error_substring_5 = "that is not equal to the measure's agg_time_dimension"
    missing_error_strings = set()
    for expected_str in [
        expected_error_substring_1,
        expected_error_substring_2,
        expected_error_substring_3,
        expected_error_substring_4,
        expected_error_substring_5,
    ]:
        if not any(actual_str.as_readable_str().find(expected_str) != -1 for actual_str in build.issues.errors):
            missing_error_strings.add(expected_str)
    assert (
        len(missing_error_strings) == 0
    ), f"Failed to match one or more expected errors: {missing_error_strings} in {set([x.as_readable_str() for x in build.issues.errors])}"


def test_count_measure_missing_expr(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests that all measures with COUNT agg should have expr provided."""
    model = copy.deepcopy(simple_model__pre_transforms)
    data_source = find_data_source_with(model, lambda ds: ds.name == "bookings_source")[0]
    invalid_measure = Measure(
        name="bad_measure",
        agg=AggregationType.COUNT,
        agg_time_dimension="ds",
    )
    data_source.measures.append(invalid_measure)  # type: ignore

    build = ModelValidator().validate_model(model)
    expected_error_substring = (
        "Measure 'bad_measure' uses a COUNT aggregation, which requires an expr to be provided. "
        "Provide 'expr: 1' if a count of all rows is desired."
    )
    assert len(build.issues.errors) == 1

    actual_error = build.issues.errors[0].as_readable_str()
    assert (
        actual_error.find(expected_error_substring) != -1
    ), f"Expected error {expected_error_substring} not found in error string! Instead got {actual_error}"


def test_count_measure_with_distinct_expr(simple_model__pre_transforms: UserConfiguredModel) -> None:
    """Tests that measures with COUNT agg can NOT use the DISTINCT keyword."""
    model = copy.deepcopy(simple_model__pre_transforms)
    data_source = find_data_source_with(model, lambda ds: ds.name == "bookings_source")[0]
    invalid_measure = Measure(
        name="distinct_count", agg=AggregationType.COUNT, agg_time_dimension="ds", expr="DISTINCT listing"
    )
    data_source.measures.append(invalid_measure)  # type: ignore

    build = ModelValidator().validate_model(model)
    expected_error_substring = "Measure 'distinct_count' uses a 'count' aggregation with a DISTINCT expr"

    assert len(build.issues.errors) == 1
    actual_error = build.issues.errors[0].as_readable_str()
    assert (
        actual_error.find(expected_error_substring) != -1
    ), f"Expected error {expected_error_substring} not found in error string! Instead got {actual_error}"



"""Helper classes to convert model objects to spec object representations

In some cases we need to take a model object and convert it to a spec for use in a DataflowPlan or similar.
This fundamentally requires us to combine a model object with a spec object and, in some cases, follow semantic
metadata about linked specs in order to resolve dimension names and the like.

Since model objects and specs should not depend on each other, we do these conversions separately via external
classes, rather than the perhaps more natural approach of adding a to_spec() method on the model objects. These
shims likely point to the need for a bit of an internal refactor, but that's a concern for another time.
"""

from typing import List

from metricflow.dataset.dataset import DataSet
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.objects.elements.measure import Measure
from metricflow.protocols.semantics import DataSourceSemanticsAccessor
from metricflow.query.query_exceptions import InvalidQueryException
from metricflow.specs import (
    DimensionSpec,
    IdentifierSpec,
    LinkableSpecSet,
    MeasureSpec,
    NonAdditiveDimensionSpec,
    SpecWhereClauseConstraint,
    StructuredLinkableSpecName,
    TimeDimensionSpec,
)


class MeasureConverter:
    """Static class for converting Measure model objects to MeasureSpec instances"""

    @staticmethod
    def convert_to_measure_spec(measure: Measure) -> MeasureSpec:
        """Converts a Measure to a MeasureSpec, and properly handles non-additive dimension properties"""
        non_additive_dimension_spec = (
            NonAdditiveDimensionSpec(
                name=measure.non_additive_dimension.name,
                window_choice=measure.non_additive_dimension.window_choice,
                window_groupings=tuple(measure.non_additive_dimension.window_groupings),
            )
            if measure.non_additive_dimension is not None
            else None
        )

        return MeasureSpec(
            element_name=measure.name,
            non_additive_dimension_spec=non_additive_dimension_spec,
        )


class WhereConstraintConverter:
    """Static class for converting WhereClauseConstraint objects to a SpecWhereClauseConstraint representation.

    The WhereClauseConstraint model object contains a parsed set of element names, and as such conversion to
    a SpecWhereClauseConstraint requires semantic resolution of linkable specs across data sources. This resolution
    has to happen for metrics and measures independently, as both can be constrained, so bolting this on
    to something like DataSourceSemantics will not be adequate.
    """

    @staticmethod
    def _convert_to_linkable_specs(
        data_source_semantics: DataSourceSemanticsAccessor, where_constraint_names: List[str]
    ) -> LinkableSpecSet:
        """Processes where_clause_constraint.linkable_names into associated LinkableInstanceSpecs (dims, times, ids)

        where_constraint_names: WhereConstraintClause.linkable_names
        data_source_semantics: DataSourceSemanticsAccessor from the instantiated class

        output: InstanceSpecSet of Tuple(DimensionSpec), Tuple(TimeDimensionSpec), Tuple(IdentifierSpec)
        """
        where_constraint_dimensions = []
        where_constraint_time_dimensions = []
        where_constraint_identifiers = []
        linkable_spec_names = [
            StructuredLinkableSpecName.from_name(linkable_name) for linkable_name in where_constraint_names
        ]
        dimension_references = {
            dimension_reference.element_name: dimension_reference
            for dimension_reference in data_source_semantics.get_dimension_references()
        }
        identifier_references = {
            identifier_reference.element_name: identifier_reference
            for identifier_reference in data_source_semantics.get_identifier_references()
        }

        for spec_name in linkable_spec_names:
            if spec_name.element_name == DataSet.metric_time_dimension_name():
                where_constraint_time_dimensions.append(TimeDimensionSpec.from_name(spec_name.qualified_name))
            elif spec_name.element_name in dimension_references:
                dimension = data_source_semantics.get_dimension(dimension_references[spec_name.element_name])
                if dimension.type == DimensionType.CATEGORICAL:
                    where_constraint_dimensions.append(DimensionSpec.from_name(spec_name.qualified_name))
                elif dimension.type == DimensionType.TIME:
                    where_constraint_time_dimensions.append(TimeDimensionSpec.from_name(spec_name.qualified_name))
                else:
                    raise RuntimeError(f"Unhandled type: {dimension.type}")
            elif spec_name.element_name in identifier_references:
                where_constraint_identifiers.append(IdentifierSpec.from_name(spec_name.qualified_name))
            else:
                raise InvalidQueryException(f"Unknown element: {spec_name}")

        return LinkableSpecSet(
            dimension_specs=tuple(where_constraint_dimensions),
            time_dimension_specs=tuple(where_constraint_time_dimensions),
            identifier_specs=tuple(where_constraint_identifiers),
        )

    @staticmethod
    def convert_to_spec_where_constraint(
        data_source_semantics: DataSourceSemanticsAccessor, where_constraint: WhereClauseConstraint
    ) -> SpecWhereClauseConstraint:
        """Converts a where constraint to one using specs."""
        return SpecWhereClauseConstraint(
            where_condition=where_constraint.where,
            linkable_names=tuple(where_constraint.linkable_names),
            linkable_spec_set=WhereConstraintConverter._convert_to_linkable_specs(
                data_source_semantics=data_source_semantics,
                where_constraint_names=where_constraint.linkable_names,
            ),
            execution_parameters=where_constraint.sql_params,
        )



from __future__ import annotations

from enum import Enum

from metricflow.object_utils import ExtendedEnum


class AggregationType(ExtendedEnum):
    """Aggregation methods for measures"""

    SUM = "sum"
    MIN = "min"
    MAX = "max"
    COUNT_DISTINCT = "count_distinct"
    # BOOLEAN is deprecated. Remove when customers have migrated.
    BOOLEAN = "boolean"
    SUM_BOOLEAN = "sum_boolean"
    AVERAGE = "average"

    COUNT = "count"
    # AVERAGE = "average"  not yet implemented ...requires us to keep track of two quantities, count and sum

    @property
    def is_expansive(self) -> bool:
        """Expansive ≝ Op( X ∪ Y ∪ ...) = Op( Op(X) ∪ Op(Y) ∪ ...)

        NOTE: COUNT is only expansive because it's transformed into a SUM agg during model transformation
        """
        return self in (
            AggregationType.SUM,
            AggregationType.MIN,
            AggregationType.MAX,
            AggregationType.BOOLEAN,
            AggregationType.COUNT,
        )

    @property
    def fill_nulls_with_0(self) -> bool:
        """Indicates if charts should show 0 instead of null where there are gaps in data."""
        return self in (
            AggregationType.SUM,
            AggregationType.COUNT_DISTINCT,
            AggregationType.SUM_BOOLEAN,
            AggregationType.COUNT,
        )

    @property
    def can_limit_dimension_values(self) -> bool:
        """Indicates if we can limit dimension values in charts.

        Currently, this means:
        1. The dimensions we care about most are the ones with the highest numeric values
        2. We can calculate the "other" column in the postprocessor (meaning the metric is expansive)
        """
        return self in (AggregationType.SUM, AggregationType.SUM_BOOLEAN, AggregationType.COUNT)


class AggregationState(Enum):
    """Represents how the measure is aggregated."""

    # When reading from the source, the measure is considered non-aggregated.
    NON_AGGREGATED = "NON_AGGREGATED"
    PARTIAL = "PARTIAL"
    COMPLETE = "COMPLETE"

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}.{self.name}"



from __future__ import annotations

import logging
import time
from abc import ABC
from contextlib import contextmanager
from typing import Iterator, Optional, Mapping, Union, Sequence, Set

import pandas as pd
import sqlalchemy

from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.base_sql_client_implementation import BaseSqlClientImplementation

logger = logging.getLogger(__name__)


class SqlAlchemySqlClient(BaseSqlClientImplementation, ABC):
    """Base class for to create DBClients for engines supported by SQLAlchemy."""

    def __init__(self, engine: sqlalchemy.engine.Engine) -> None:  # noqa: D
        self._engine = engine

    @staticmethod
    def build_engine_url(  # noqa: D
        dialect: str,
        database: str,
        username: str,
        password: Optional[str],
        host: str,
        port: Optional[int] = None,
        query: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
        driver: Optional[str] = None,
    ) -> sqlalchemy.engine.url.URL:
        return sqlalchemy.engine.url.URL.create(
            f"{dialect}+{driver}" if driver else f"{dialect}",
            username=username,
            password=password,
            host=host,
            port=port,
            database=database,
            **({"query": query} if query is not None else {}),
        )

    @staticmethod
    def create_engine(  # noqa: D
        dialect: str,
        port: int,
        database: str,
        username: str,
        password: str,
        host: str,
        driver: Optional[str] = None,
        query: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
    ) -> sqlalchemy.engine.Engine:

        connect_url = SqlAlchemySqlClient.build_engine_url(
            dialect=dialect,
            driver=driver,
            username=username,
            password=password,
            host=host,
            port=port,
            database=database,
            query=query,
        )
        # Without pool_pre_ping, it's possible for timed-out connections to be returned to the client and cause errors.
        # However, this can cause increase latency for slow engines.
        return sqlalchemy.create_engine(
            connect_url,
            pool_size=10,
            max_overflow=10,
            pool_pre_ping=True,
        )

    def list_tables(self, schema_name: str) -> Sequence[str]:  # noqa: D
        return self._engine.table_names(schema=schema_name)

    @contextmanager
    def engine_connection(self, engine: sqlalchemy.engine.Engine) -> Iterator[sqlalchemy.engine.Connection]:
        """Context Manager for providing a configured connection."""
        conn = engine.connect()
        try:
            yield conn
        finally:
            conn.close()

    def _engine_specific_query_implementation(  # noqa: D
        self, stmt: str, bind_params: SqlBindParameters
    ) -> pd.DataFrame:
        with self.engine_connection(self._engine) as conn:
            return pd.read_sql_query(sqlalchemy.text(stmt), conn, params=bind_params.param_dict)

    def _engine_specific_execute_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:  # noqa: D
        with self.engine_connection(self._engine) as conn:
            conn.execute(sqlalchemy.text(stmt), bind_params.param_dict)

    def _engine_specific_dry_run_implementation(self, stmt: str, bind_params: SqlBindParameters) -> None:  # noqa: D
        with self.engine_connection(self._engine) as conn:
            s = "EXPLAIN " + stmt
            conn.execute(sqlalchemy.text(s), bind_params.param_dict)

    def create_table_from_dataframe(  # noqa: D
        self, sql_table: SqlTable, df: pd.DataFrame, chunk_size: Optional[int] = None
    ) -> None:
        logger.info(f"Creating table '{sql_table.sql}' from a DataFrame with {df.shape[0]} row(s)")
        start_time = time.time()
        with self.engine_connection(self._engine) as conn:
            pd.io.sql.to_sql(
                frame=df,
                name=sql_table.table_name,
                con=conn,
                schema=sql_table.schema_name,
                index=False,
                if_exists="fail",
                method="multi",
                chunksize=chunk_size,
            )
        logger.info(f"Created table '{sql_table.sql}' from a DataFrame in {time.time() - start_time:.2f}s")

    @staticmethod
    def validate_query_params(
        url: sqlalchemy.engine.url.URL,
        required_parameters: Set[str],
        optional_parameters: Set[str],
    ) -> None:
        """Checks that the query parameters in the URL only include the valid parameters specified."""
        query_keys = set(url.query.keys())
        errors = []
        if not query_keys.issuperset(required_parameters):
            errors.append(f"Missing required parameters {required_parameters - query_keys}")
        if not query_keys.issubset(required_parameters.union(optional_parameters)):
            errors.append(f"Found extra parameters {query_keys - required_parameters.union(optional_parameters)}")

        if errors:
            raise ValueError(f"Found errors in the URL: {url}\n" + "\n".join(errors))



import os

from _pytest.fixtures import FixtureRequest

from metricflow.dag.dag_visualization import DagGraphT, render_via_graphviz
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.plan_utils import snapshot_path_prefix


def display_graph_if_requested(
    mf_test_session_state: MetricFlowTestSessionState,
    request: FixtureRequest,
    dag_graph: DagGraphT,
) -> None:
    """Create and display the plan as an SVG, if requested to do so."""

    if not mf_test_session_state.display_plans:
        return

    plan_svg_output_path_prefix = snapshot_path_prefix(
        request=request, snapshot_group=dag_graph.__class__.__name__, snapshot_id=dag_graph.dag_id
    )

    # Create parent directory since it might not exist
    os.makedirs(os.path.dirname(plan_svg_output_path_prefix), exist_ok=True)

    if mf_test_session_state.plans_displayed >= mf_test_session_state.max_plans_displayed:
        raise RuntimeError(
            f"Can't display plan - hit limit of {mf_test_session_state.max_plans_displayed} plans displayed."
        )
    render_via_graphviz(dag_graph=dag_graph, file_path_without_svg_suffix=plan_svg_output_path_prefix)
    mf_test_session_state.plans_displayed += 1



from metricflow.dataflow.builder.measure_additiveness import group_measure_specs_by_additiveness
from metricflow.aggregation_properties import AggregationType
from metricflow.specs import MeasureSpec, NonAdditiveDimensionSpec


def test_bucket_measure_specs_by_additiveness() -> None:  # noqa: D
    # Semi-additive Bucket 1
    measure_1 = MeasureSpec(
        element_name="measure_1",
        non_additive_dimension_spec=NonAdditiveDimensionSpec(
            name="ds",
            window_choice=AggregationType.MIN,
        ),
    )

    # Semi-additive Bucket 2
    measure_2 = MeasureSpec(
        element_name="measure_2",
        non_additive_dimension_spec=NonAdditiveDimensionSpec(
            name="ds",
            window_choice=AggregationType.MIN,
            window_groupings=("id_1", "id_2"),
        ),
    )
    measure_3 = MeasureSpec(
        element_name="measure_3",
        non_additive_dimension_spec=NonAdditiveDimensionSpec(
            name="ds",
            window_choice=AggregationType.MIN,
            window_groupings=("id_2", "id_1"),
        ),
    )

    # Additive Bucket
    measure_4 = MeasureSpec(element_name="measure_4")

    result = group_measure_specs_by_additiveness((measure_1, measure_2, measure_3, measure_4))
    assert result.additive_measures == (measure_4,)
    assert (measure_1,) in result.grouped_semi_additive_measures
    assert (measure_2, measure_3) in result.grouped_semi_additive_measures
